from typing import List, Tuple, Dict, Any
from collections import namedtuple
from enum import IntEnum


# ------------------------------------------------------------------------------
#           WindowMappingType
# ------------------------------------------------------------------------------
class WindowMappingType(IntEnum):

    """
    An integer enumertaion used to identify the different windows used in the Level 1 data processing

    Attributes
    ----------
    WINDOW_UNDEFINED: int(1)
        Undefined window coordinates. This is acceptable when it is not critical to track the exact window definition. This is the default.
    WINDOW_NOOF: int(2)
        Identifies coordinates in the near out-of-field window
    WINDOW_DETECTOR: int(3)
        Identifies coordinates in the detector window
    WINDOW_SCIENCE: int(4)
        Identifies coordinates in Science window pixels
    WINDOW_L1B: int(5)
        Identifies coordinates in L1B pixels
    WINDOW_L1C: int(6)
        Identifies coordinates in L1C pixels.
    WINDOW_FULLCONV: int(7)
        Identifies window for full convolution
    RADIANS : int(8)
        Identifies coordinates measured in radians
    """
    WINDOW_UNDEFINED = 1
    WINDOW_NOOF = 2
    WINDOW_DETECTOR = 3
    WINDOW_SCIENCE = 4
    WINDOW_L1B = 5
    WINDOW_L1C = 6
    WINDOW_FULLCONV = 7
    RADIANS = 8


# ------------------------------------------------------------------------------
#           AltiusWindowSpecs
# ------------------------------------------------------------------------------
#: Attributes
#: ----------
#: NOFSx: float
#:     The X offset of the bottom left corner of the NOOF window in FULLCONV window coordinates.
#: NOFSy: float
#:     The Y offset of the bottom left corner of the NOOF window in FULLCONV window coordinates.
#: NOOFx: float
#:     The horizontal size of the near out-of-field window in detector pixels.
#: NOOFy: float
#:     The vertical size of the near out-of-field window in detector pixels
#: DOFSx: float
#:     The X offset of the bottom left corner of the detector window in NOOF window coordinates.
#: DOFSy: float
#:     The Y offset of the bottom left corner of the detector window in NOOF window coordinates.
#: Dx: float
#:     The horizontal size of the detector window in detector pixels, eg 1504
#: Dy: float
#:     The vertical size of the detector window in detector pixels, e.g. 2000
#: SOFSx: float
#:     The X offset of the bottom left corner of the science window in detector window coordinates/pixels. This origin is
#:     implicitly shared with the L1B and L1C windows.
#: SOFSy: float
#:     The Y offset of the bottom left corner of the science window in detector window coordinates/pixels.  This origin is
#:     implicitly shared with the L1B and L1C windows.
#: Sx: float
#:     The horizontal size of the science window in detector/science pixels, eg 1500. This physical size is
#:     implicitly shared with the L1B and L1C windows.
#: Sy: float
#:     The vertical size of the science window in detector/science pixels, e.g. 1503. This physical size is
#:     implicitly shared with the L1B and L1C windows.
#: LBINx: float
#:     The horizontal binning of L1B pixels in science pixels, e.g. 3. The value of Sx must be an exact multiple of this value.
#: LBINy: float
#:     The vertical binning of L1B pixels in science pixels, e.g. 3. The value of Sy must be an exact multiple of this value.
#: Lx: float
#:     The horizontal size of the L1B window in L1B pixels, eg 500. Lx * LBINx must equal Sx
#: Ly: float
#:     The vertical size of the L1B window in L1B pixels, eg 501. Ly * LBINy must equal Sy
#: KBINx: float
#:     The horizontal binning of L1C pixels in L1B pixels, e.g. 50. The value of Lx must be an exact multiple of this value.
#: KBINy: float
#:     The vertical binning of L1C pixels in L1B pixels, e.g. 1. The value of Ly must be an exact multiple of this value.
#: Kx: float
#:     The horizontal size of the L1C window in L1C pixels, eg 10. Kx * KBINx must equal Lx
#: Ky: float
#:     The vertical size of the L1C window in L1C pixels, eg 501. Ky * KBINy must equal Ly
AltiusWindowSpecs = namedtuple('AltiusWindowSpecs', ['FULLCONVx', 'FULLCONVy',
                                                     'NOFSx', 'NOFSy', 'NOOFx', 'NOOFy',
                                                     'DOFSx', 'DOFSy', 'Dx', 'Dy',
                                                     'SOFSx', 'SOFSy', 'Sx', 'Sy',
                                                     'LBINx', 'LBINy', 'Lx', 'Ly',
                                                     'KBINx', 'KBINy', 'Kx', 'Ky'])


# ------------------------------------------------------------------------------
#           Coords
# ------------------------------------------------------------------------------
#: Attributes
#: ----------
#: x: float
#:     The horizontal coordinate.
#: y: float
#:     The vertical coordinate
#: windowtype: WindowMappingType
#:    Defines the window in which the coordinates are defined.
Coords = namedtuple('Coords', ['y', 'x', 'windowtype'], defaults=(('y', 0), ('x', 0), ('windowtype', WindowMappingType.WINDOW_UNDEFINED)))


# ------------------------------------------------------------------------------
#           coord_from_tuple
# ------------------------------------------------------------------------------
def coord_from_tuple(l: Tuple[float, float, int]):
    """
    Converst a 3 tuple element into Coord object.
    """

    wintype = int(l[2])
    window = WindowMappingType.WINDOW_UNDEFINED
    if (wintype == WindowMappingType.WINDOW_NOOF):
        window = WindowMappingType.WINDOW_NOOF
    elif (wintype == WindowMappingType.WINDOW_DETECTOR):
        window = WindowMappingType.WINDOW_DETECTOR
    elif (wintype == WindowMappingType.WINDOW_SCIENCE):
        window = WindowMappingType.WINDOW_SCIENCE
    elif (wintype == WindowMappingType.WINDOW_L1B):
        window = WindowMappingType.WINDOW_L1B
    elif (wintype == WindowMappingType.WINDOW_L1C):
        window = WindowMappingType.WINDOW_L1C
    elif (wintype == WindowMappingType.WINDOW_FULLCONV):
        window = WindowMappingType.WINDOW_FULLCONV
    elif (wintype == WindowMappingType.RADIANS):
        window = WindowMappingType.RADIANS
    coords = Coords(y=l[0], x=l[1], windowtype=window)
    return coords


# ------------------------------------------------------------------------------
#           altius_window_specs_as_list
# ------------------------------------------------------------------------------
def altius_window_specs_as_list(window_mapping: AltiusWindowSpecs) -> List[int]:
    """
    Converts an instance of AltiusWindowSpecs into a list. This list is compatible with the list of values required by
    the C++ deconvolution matrix code and its order should not be changed.

    Parameters
    ----------
    window_mapping: AltiusWindowSpecs
        The namedtuple containing all of the detector window definitions used by the Altius instrument

    Returns
    -------
    A list of 18 integers that can be passed to the L1C deconvolution matrix code.
    """

    array_window_specs = [window_mapping.FULLCONVx, window_mapping.FULLCONVy,
                          window_mapping.NOFSx, window_mapping.NOFSy, window_mapping.NOOFx, window_mapping.NOOFy,
                          window_mapping.DOFSx, window_mapping.DOFSy, window_mapping.Dx, window_mapping.Dy,
                          window_mapping.SOFSx, window_mapping.SOFSy, window_mapping.Sx, window_mapping.Sy,
                          window_mapping.LBINx, window_mapping.LBINy, window_mapping.Lx, window_mapping.Ly,
                          window_mapping.KBINx, window_mapping.KBINy, window_mapping.Kx, window_mapping.Ky]
    return array_window_specs


# ------------------------------------------------------------------------------
#           altius_window_specs_from_dict
# ------------------------------------------------------------------------------
def altius_window_specs_from_dict(window_mapping: Dict[str, Any]) -> AltiusWindowSpecs:
    """
    Converts a dictonary of window specs into a namedtuple of AltiusWindowSpecs. This is a helper function
    for code reading window specs from netcdf files. The PSF I/O code uses it.
    """

    spec = AltiusWindowSpecs(FULLCONVx=window_mapping.get('FULLCONVx'),
                             FULLCONVy=window_mapping.get('FULLCONVy'),
                             NOFSx=window_mapping.get('NOFSx'),
                             NOFSy=window_mapping.get('NOFSy'),
                             NOOFx=window_mapping['NOOFx'],
                             NOOFy=window_mapping['NOOFy'],
                             DOFSx=window_mapping['DOFSx'],
                             DOFSy=window_mapping['DOFSy'],
                             Dx=window_mapping['Dx'],
                             Dy=window_mapping['Dy'],
                             SOFSx=window_mapping['SOFSx'],
                             SOFSy=window_mapping['SOFSy'],
                             Sx=window_mapping['Sx'],
                             Sy=window_mapping['Sy'],
                             LBINx=window_mapping['LBINx'],
                             LBINy=window_mapping['LBINy'],
                             Lx=window_mapping['Lx'],
                             Ly=window_mapping['Ly'],
                             KBINx=window_mapping['KBINx'],
                             KBINy=window_mapping['KBINy'],
                             Kx=window_mapping['Kx'],
                             Ky=window_mapping['Ky'])
    return spec


# ------------------------------------------------------------------------------
#           Class WindowMapping
# ------------------------------------------------------------------------------
class WindowMapping:
    '''
    Object that creates linear mappings of offset and binning from one window to another. This is useful for
    managing the different windows (on the detector surface) used in the ALTIUS project.
    '''

    def __init__(self,
                 windowtype: WindowMappingType,
                 offset=Coords(y=0, x=0, windowtype=WindowMappingType.WINDOW_UNDEFINED),
                 binning=Coords(y=1, x=1, windowtype=WindowMappingType.WINDOW_UNDEFINED),
                 size=Coords(y=0, x=0, windowtype=WindowMappingType.WINDOW_UNDEFINED),
                 parent: 'WindowMapping' = None):

        self._windowtype = windowtype
        self._parent: 'WindowMapping' = parent
        self._offset: Coords = offset
        self._binning: Coords = binning
        self._size: Coords = size

    # ------------------------------------------------------------------------------
    #               def CreateMappingToUpperLevel
    # ------------------------------------------------------------------------------
    def CreateMappingToUpperLevel(self, upperlevelparent: 'WindowMapping') -> 'WindowMapping':
        '''
        Walks up a window hierarchy to make a mapping from a lower window to a higher window.

        Parameters
        ----------
        upperlevelparent: WindowMapping
            The upper level parent in the hierarchy. This object must be the current window in the window hierarchy

        Returns
        -------
        WindowMapping
            The mapping from the current window to the upper parent window.
        '''

        ofsx = 0
        ofsy = 0
        binx = 1
        biny = 1
        current_window = self

        while ((current_window is not upperlevelparent) and (current_window is not None)):
            ofsx = ofsx * current_window._binning.x + current_window._offset.x
            ofsy = ofsy * current_window._binning.y + current_window._offset.y
            binx *= current_window._binning.x
            biny *= current_window._binning.y
            current_window = current_window._parent

        total_offset = Coords(y=ofsy, x=ofsx, windowtype=upperlevelparent._windowtype)
        total_binning = Coords(y=biny, x=binx, windowtype=upperlevelparent._windowtype)

        assert current_window is upperlevelparent
        newmapping = WindowMapping(self._windowtype, offset=total_offset, binning=total_binning, size=self._size, parent=upperlevelparent)
        return newmapping

    # ------------------------------------------------------------------------------
    #           ToParentWindow
    # ------------------------------------------------------------------------------
    def ToParentWindow(self, sourcecoords: Coords, checkrange: bool = True) -> Coords:

        assert self._parent is not None
        x = sourcecoords.x * self._binning.x + self._offset.x
        y = sourcecoords.y * self._binning.y + self._offset.y
        parentcoords = Coords(x=x, y=y, windowtype=self._parent._windowtype)
        if (checkrange):
            ok1 = (parentcoords.x < self._parent._size.x) and (parentcoords.y < self._parent._size.y) and (parentcoords.x >= 0) and (parentcoords.y >= 0)
            assert(ok1)
        return parentcoords

    # ------------------------------------------------------------------------------
    #           FromParentWindow
    # ------------------------------------------------------------------------------
    def FromParentWindow(self, parentcoords, checkrange: bool = True) -> Coords:

        x = (parentcoords.x - self._offset.x) / self._binning.x
        y = (parentcoords.y - self._offset.y) / self._binning.y
        sourcecoords = Coords(x=x, y=y, windowtype=self._windowtype)
        if (checkrange):
            ok = (sourcecoords.x < self._size.x) and (sourcecoords.y < self._size.y) and (sourcecoords.x >= 0) and (sourcecoords.y >= 0)
            assert ok
        return sourcecoords

    def Numx(self):
        return self._size.x

    def Numy(self):
        return self._size.y

    def Binx(self):
        return self._binning.x

    def Biny(self):
        return self._binning.y

    def Ofsx(self):
        return self._offset.x

    def Ofsy(self):
        return self._offset.y


# ------------------------------------------------------------------------------
#           WindowMappingHierarchy
# ------------------------------------------------------------------------------
class WindowMappingHierarchy:

    def __init__(self, specs: AltiusWindowSpecs):

        self.fullconv_window = WindowMapping(WindowMappingType.WINDOW_FULLCONV,
                                             size=Coords(y=specs.FULLCONVy, x=specs.FULLCONVx, windowtype=WindowMappingType.WINDOW_FULLCONV),
                                             parent=None)
        self.noof_window = WindowMapping(WindowMappingType.WINDOW_NOOF,
                                         offset=Coords(y=specs.NOFSy, x=specs.NOFSx, windowtype=WindowMappingType.WINDOW_FULLCONV),
                                         binning=Coords(y=1, x=1, windowtype=WindowMappingType.WINDOW_FULLCONV),
                                         size=Coords(y=specs.NOOFy, x=specs.NOOFx, windowtype=WindowMappingType.WINDOW_NOOF),
                                         parent=self.fullconv_window)
        self.detector_window = WindowMapping(WindowMappingType.WINDOW_DETECTOR,
                                             offset=Coords(y=specs.DOFSy, x=specs.DOFSx, windowtype=WindowMappingType.WINDOW_NOOF),
                                             binning=Coords(y=1, x=1, windowtype=WindowMappingType.WINDOW_NOOF),
                                             size=Coords(y=specs.Dy, x=specs.Dx, windowtype=WindowMappingType.WINDOW_DETECTOR),
                                             parent=self.noof_window)
        self.sci_window = WindowMapping(WindowMappingType.WINDOW_SCIENCE,
                                        offset=Coords(y=specs.SOFSy, x=specs.SOFSx, windowtype=WindowMappingType.WINDOW_DETECTOR),
                                        binning=Coords(y=1, x=1, windowtype=WindowMappingType.WINDOW_DETECTOR),
                                        size=Coords(y=specs.Sy, x=specs.Sx, windowtype=WindowMappingType.WINDOW_SCIENCE),
                                        parent=self.detector_window)
        self.l1b_window = WindowMapping(WindowMappingType.WINDOW_L1B,
                                        offset=Coords(y=0, x=0, windowtype=WindowMappingType.WINDOW_SCIENCE),
                                        binning=Coords(y=specs.LBINy, x=specs.LBINx, windowtype=WindowMappingType.WINDOW_SCIENCE),
                                        size=Coords(y=specs.Ly, x=specs.Lx, windowtype=WindowMappingType.WINDOW_L1B),
                                        parent=self.sci_window)
        self.l1c_window = WindowMapping(WindowMappingType.WINDOW_L1C,
                                        binning=Coords(y=specs.KBINy, x=specs.KBINx, windowtype=WindowMappingType.WINDOW_L1B),
                                        offset=Coords(y=0, x=0, windowtype=WindowMappingType.WINDOW_L1B),
                                        size=Coords(y=specs.Ky, x=specs.Kx, windowtype=WindowMappingType.WINDOW_L1C),
                                        parent=self.l1b_window)
        self.l1ctodetector = self.l1c_window.CreateMappingToUpperLevel(self.detector_window)

        self._windows = {WindowMappingType.WINDOW_FULLCONV: self.fullconv_window,
                         WindowMappingType.WINDOW_NOOF: self.noof_window,
                         WindowMappingType.WINDOW_DETECTOR: self.detector_window,
                         WindowMappingType.WINDOW_SCIENCE: self.sci_window,
                         WindowMappingType.WINDOW_L1B: self.l1b_window,
                         WindowMappingType.WINDOW_L1C: self.l1c_window}

    @property
    def NOOF_larger_than_detector(self) -> bool:
        """
        True if the NOOF window is larger than the detector window. This is used to indicate if modified PSF or measured PSF
        can be used in calculations that require the PSF across the entire NOOF region.
        :return:
        """

        return (self.noof_window.Numx() > self.detector_window.Numx()) or (self.noof_window.Numy() > self.detector_window.Numy())

    # ------------------------------------------------------------------------------
    #           __getitem__
    # ------------------------------------------------------------------------------
    def __getitem__(self, windowtype: WindowMappingType) -> WindowMapping:
        """Fetch the window mapping for the requested window type"""
        return self._windows[windowtype]

    # ------------------------------------------------------------------------------
    #           convert_to
    # ------------------------------------------------------------------------------
    def convert_to(self, location: Coords, destination_type: WindowMappingType, checkrange: bool=False):
        """
        Converts coordinates expressed in one window to coordinates expressed in another window.
        This works for any window to any other window. It is useful for converting to the detector window as an intermediate
        step for converting to angle.

        Parameters
        ----------
        location: Coords
            The x,y location in teh source window
        destination_type: WindowMappingType
            The destination window type.  Any WindowMappingType except RADIANS or WINDOW_UNDEFINED.
        checkrange: bool
            If True then check destination coordinates are within the bounds of the destination window. Default is False

        Returns
        -------
        Coords
            The coordinates of the locationm in the destination window.
        """

        source_type = location.windowtype
        sourcemap = self._windows[source_type]
        destmap = self._windows[destination_type]

        if destination_type < source_type:
            map = sourcemap.CreateMappingToUpperLevel(destmap)
            newcoords = map.ToParentWindow(location, checkrange=checkrange)
        elif source_type < destination_type:                                            # if the source is higher up the hierarchy it is the parent
            map = destmap.CreateMappingToUpperLevel(sourcemap)                          # so create the mapping from the destination window
            newcoords = map.FromParentWindow(location, checkrange=checkrange)           # and convert from
        else:
            assert(source_type == destination_type)
            newcoords = location
        return newcoords
