from typing import Tuple, Dict, List
import numpy as np
from l1c_deconvolution_matrix import generate_convolution_matrix, generate_horizontal_stray_light_matrix
from l1c_deconvolution_matrix.matrix_file_io import write_to_netcdf, load_from_netcdf
from ..configuration import AltiusWindowSpecs, WindowMappingType, Coords
from ..configuration.window_mapping import altius_window_specs_as_list


# ------------------------------------------------------------------------------
#           ConvolutionMatrix
# ------------------------------------------------------------------------------
class L1CConvolutionMatrix:
    """
    A class to manage the convolution matrices used by L1C processing.  The convolution matrices are large arrays that
    are stored as 2-D arrays but are really 4-D arrays which store a 2-D array of values for every location in a 2-D array.
    The rows of the matrix provide the relationship between the blurred, measured signal, :math:`Y_i` at a given image location
    and the contribution to that signal from the true input signal at all other locatuons in the image, :math:`\sum_j A_{i,j} X_j`.
    The rows and columns of the matrix are both represented as 1-D array of indices but in actuality map to 2-D pixel locations.
    The indexing is performed in a column major manner with the 1-D index of a pixel at 2-D location (y,x) given by
    :math:`x\,N_y + y` where :math:`N_y` is the number of pixels in the 2-D column.
    """

    def __init__(self, a_matrix: np.ndarray = None, window_mapping: AltiusWindowSpecs = None):

        self._a_matrix: np.ndarray = None
        self._window_mapping: AltiusWindowSpecs = None
        self._Kx = 0
        self._Ky = 0
        self._KK = 0
        if (a_matrix is not None):
            self.set_matrix(a_matrix, window_mapping)

    @property
    def matrix(self) -> np.ndarray:
        """
        The internal numpy array representing the matrix.
        """
        return self._a_matrix

    # ------------------------------------------------------------------------------
    #           Inverse
    # ------------------------------------------------------------------------------

    def Inverse(self) -> 'L1CConvolutionMatrix':
        """
        Calculates the matrix inverse of the internal numpy array
        """
        inv = np.linalg.inv(self._a_matrix)
        return L1CConvolutionMatrix(inv, self._window_mapping)

    # ------------------------------------------------------------------------------
    #           set_matrix
    # ------------------------------------------------------------------------------

    def set_matrix(self, a_matrix: np.ndarray, window_mapping: AltiusWindowSpecs):
        """
        Sets the L1C matrix to the given array. The user must also provide the window mapping so the class knows how
        the linear array breaks down into 2-D rows and columns.

        Parameters
        ----------
        a_matrix: np.ndarray[KK, KK]
            The L1C convolution (or deconvolution) matrix. It must be a square array of dimension ``[KK, KK]`` where
            ``KK`` is the product of the L1C number of rows (Ny) and number of columns(Nx)

        window_mapping: AltiusWindowSpecs
            The Altius window specifications associated with thegiven convolution matrix.

        """
        self._a_matrix = a_matrix
        self._window_mapping = window_mapping
        self._Kx = window_mapping.Kx
        self._Ky = window_mapping.Ky
        self._KK = self._Kx * self._Ky
        s = a_matrix.shape
        assert (s[0] == self._KK) and (s[1] == self._KK), "The incoming convolution matrix does not match teh expected size."

    # ------------------------------------------------------------------------------
    #           __add__
    # ------------------------------------------------------------------------------
    def __add__(self, other) -> 'L1CConvolutionMatrix':
        """
        Implements matrix addition
        :param other:
        :return:
        """
        m = self._a_matrix + other._a_matrix
        answer = L1CConvolutionMatrix(m, self._window_mapping)
        return answer

    # ------------------------------------------------------------------------------
    #           __matmul__
    # ------------------------------------------------------------------------------
    def __matmul__(self, l1c: np.ndarray) -> np.ndarray:
        """
        Implements the matrix multiplication operator @ for the L1C convolution matrix. This operator overload allows
        the user to multiply with either a regular 2-D L1C matrix or a flattened L1C matrix. In either case the incoming array
        is flattened to a column vector, multiplied and then returned to the same format as the incoming arrayu.

        Parameters
        ----------
        l1c : np.ndarray[Ky, Kx]
            A 2-D array of L1C data product. The first dimension, Ky,  is the Y (vertical) axis and the second, Kx, is the X axis.
            Alternatively it can be a 1D array of size Kx * Ky, in which case the user must ensure the elements are in the correct
            column major ordering: Ky varies rapidly and Kx varies slowly

        Returns
        -------
        np.ndarray[Ky,Kx]
            Returns an array the same size and shape as the incoming ``l1c`` array that is the matrix product of the
            internal convolution matrix and the incoming l1c array.
        """
        s = l1c.shape
        is2d = (len(s) == 2)
        if (is2d):
            assert (s[0] == self._Ky) and (s[1] == self._Kx), "The incoming L1c matrix does not match the convolution matrix size"
            columnvector = l1c.transpose().flatten()
        else:
            assert (len(s) == 1) and (s[0] == self._KK)
            columnvector = l1c
        answer = self._a_matrix @ columnvector
        if (is2d):
            answer = answer.reshape([self._Kx, self._Ky]).transpose()
        return answer

    # ------------------------------------------------------------------------------
    #           load_matrix_file
    # ------------------------------------------------------------------------------
    @classmethod
    def load_matrix_file(cls, filename: str) -> Tuple['L1CConvolutionMatrix', AltiusWindowSpecs, Coords]:
        """
        Loads the convolution matrix from a netcdf file. This is a shallow wrapper for the actual netcdf I/O code stored
        in package L1C_deconvolution_matrix. It is assumed this convolution matrix was written using
        method ``L1C_deconvolution_matrix.write_to_netcdf``

        Parameters
        ----------
        filename: str
            The name of the netcdf file. It must exist, be accessible and be a netcdf file created by method ``write_to_netcdf``.

        Returns
        -------
        Tuple[ convolution_matrix, window_specs, psf_location_array]
            Returns a four element tuple. The first element is the convolution matrix stored in the netcdf file. It is returned
            as a numpy array. The 2nd element is a dictionary listing the window specifications used to create the convolution
            matrix. The third element is a  numpy array of dimensions (3, Numpsf) that specify the (Y,X,) location of the central lobe of
            each PSF and the window type of the coordinate (2 = NOOF window, 3 = detector window, 4 = Science window).
        """

        matrix, windowspecs, center = load_from_netcdf(filename)
        amatrix = L1CConvolutionMatrix(matrix)
        acenter = Coords(y=center[0], x=center[1], windowtype=int(center[2]))
        windows = AltiusWindowSpecs(FULLCONVx=windowspecs['FULLCONVx'],
                                    FULLCONVy=windowspecs['FULLCONVy'],
                                    NOOFx=windowspecs['NOOFx'],
                                    NOOFy=windowspecs['NOOFy'],
                                    DOFSx=windowspecs['DOFSx'],
                                    DOFSy=windowspecs['DOFSy'],
                                    Dx=windowspecs['Dx'],
                                    Dy=windowspecs['Dy'],
                                    SOFSx=windowspecs['SOFSx'],
                                    SOFSy=windowspecs['SOFSy'],
                                    Sx=windowspecs['Sx'],
                                    Sy=windowspecs['Sy'],
                                    LBINx=windowspecs['LBINx'],
                                    LBINy=windowspecs['LBINy'],
                                    Lx=windowspecs['Lx'],
                                    Ly=windowspecs['Ly'],
                                    KBINx=windowspecs['KBINx'],
                                    KBINy=windowspecs['KBINy'],
                                    Kx=windowspecs['Kx'],
                                    Ky=windowspecs['Ky'])
        return amatrix, windows, acenter
