from typing import Tuple
from collections import namedtuple
import numpy as np
import matplotlib.pyplot as plt
from ..configuration import Coords, OpticsSpecs, PSFSpecs, PSFSpecsWavelengthEntry, WindowMappingType, WindowMappingHierarchy
from ..configuration.psf_configuration import HarveyScatteringCoeffs
from ..psf import PSFTools


class ModifiedHarveyPSF:

    # ------------------------------------------------------------------------------
    #           modified_harvey_stray_light_bsdf
    # ------------------------------------------------------------------------------
    @classmethod
    def modified_harvey_stray_light_bsdf(cls, angles: np.ndarray, coeffs: HarveyScatteringCoeffs) -> np.ndarray:
        """
        Calculates the stray light signal as a function of angle using the modified harvey function, :math:`B_0\\left[1+\\left(\\frac{\\theta}{l}\\right)^2\\right]^{-\\frac{S}{2}}`.

        Parameters
        ----------
        angles: np.ndarray[N]
            An array of angles measured in radians from the on-axis location. The modified harvey stray light is calculated at each angle.

        coeffs: HarveyScatteringCoeffs
            A named typle containing the b0, l and S used to parameterise the Modified Harvey function

        Returns
        -------
        np.ndarray[N]
            Returns the value of the modified harvey PSF as an array. It is the same size as the input angles array.
        """
        bsdf = np.power(1.0 + np.square(angles / coeffs.l), coeffs.S / -2.0) * coeffs.b0
        return bsdf

    # ------------------------------------------------------------------------------
    #           fit_modified_harvey_curve_to_vis
    # ------------------------------------------------------------------------------
    @classmethod
    def fit_modified_harvey_curve_to_psf(cls,
                                         specs: OpticsSpecs,
                                         psfspecs: PSFSpecsWavelengthEntry,
                                         psf: np.ndarray,
                                         center: Coords,
                                         fignum: int = None) -> HarveyScatteringCoeffs:
        """
        Fits a modified harvey point spread function to the measured VIS psf.
        This currently only fits in the horizontal direction and is manually adjusted
        to fit to the regions of interest
        :return:
        """

        rads_per_pix = specs.radians_per_detector_pixel
        r = PSFTools.get_radial_distance_of_pixels(psf, center)
        angle = r * rads_per_pix
        anglerange = psfspecs.far_field_angular_fit_range
        good = (psf > 0.0)
        valid = np.zeros_like(good)
        for i in range(len(anglerange) // 2):
            j = 2 * i
            valid = valid | ((angle > anglerange[j]) & (angle < anglerange[j + 1]))
        good = good & valid

        numbadspots = len(psfspecs._far_field_bad_spots) // 3
        for i in range(numbadspots):
            j = 3 * i
            spot_y = psfspecs._far_field_bad_spots[j]
            spot_x = psfspecs._far_field_bad_spots[j + 1]
            spot_radius = psfspecs._far_field_bad_spots[j + 2]
            spot_center = Coords(y=spot_y, x=spot_x, windowtype=WindowMappingType.WINDOW_DETECTOR)
            rspot = PSFTools.get_radial_distance_of_pixels(psf, spot_center)
            goodspot = rspot > spot_radius
            good = good & goodspot

        x = np.log(angle[good])
        y = np.log(psf[good])
        coeffs = np.polyfit(x, y, 1)
        xt = np.linspace(np.min(x), np.max(x), 1000)
        yt = np.polyval(coeffs, xt)
        S = -coeffs[0]
        l = 0.001
        bb = coeffs[1] - S * np.log(l)
        b0 = np.exp(bb)
        far_field_harvey_coeffs = HarveyScatteringCoeffs(S=S, l=l, b0=b0)
        if (fignum is not None):
            angle_theory = np.linspace(0.0, np.max(angle[good]), 1000)
            theory = cls.modified_harvey_stray_light_bsdf(angle_theory, far_field_harvey_coeffs)
            plt.figure(fignum)
            plt.clf()
            plt.plot((angle[good].flatten()), np.log(psf[good].flatten()), 'k.', label='All Radii', markersize=0.1)
            plt.plot((angle[:, 752]), np.log(psf[:, 752]), 'ro', label='Vertical Slice')
            plt.plot((angle[1000, :]), np.log(psf[1000, :]), 'gs', label='Horizontal Slice')
            plt.plot(angle_theory, np.log(theory), 'm-', label='Fitted Harvey curve')
            plt.plot(np.exp(xt), yt, 'c-', label='Fitted power curve')
            plt.xlabel('Angle from center in radians')
            plt.ylabel('Ln(PSF)')
            plt.legend()
            plt.title('Radial distribution of {} PSF'.format(specs.channel))
            plt.figure(fignum + 1)
            maskedpsf = np.array(psf)
            bad = ~good
            maskedpsf[bad] = 0
            plt.imshow(np.log10(maskedpsf), origin='lower')
            plt.colorbar()
        # farfield_theory = modified_harvey_stray_light_bsdf( angle, far_field_harvey_coeffs)
        return far_field_harvey_coeffs

    # ------------------------------------------------------------------------------
    #           fit_modified_harvey_curve_to_vis
    # ------------------------------------------------------------------------------
    @classmethod
    def fullconv_image_from_modified_harvey_curve(cls,
                                                  specs: OpticsSpecs,
                                                  far_field_harvey_coeffs: HarveyScatteringCoeffs) -> Tuple[np.ndarray, Coords]:
        """
        Fits a modified harvey point spread function to the measured VIS psf.
        This currently only fits in the horizontal direction and is manually adjusted
        to fit to the regions of interest
        :return:
        """

        fullconv_window = specs.detector_specs.window_hierarchy.fullconv_window
        Nx = fullconv_window.Numx()
        Ny = fullconv_window.Numy()
        rads_per_pix = specs.radians_per_detector_pixel
        psf, center = cls.image_from_modified_harvey_curve(Ny, Nx, rads_per_pix, WindowMappingType.WINDOW_FULLCONV, far_field_harvey_coeffs)
        return psf, center

    # ------------------------------------------------------------------------------
    #           fit_modified_harvey_curve_to_vis
    # ------------------------------------------------------------------------------
    @classmethod
    def image_from_modified_harvey_curve(cls,
                                         Ny: int,
                                         Nx: int,
                                         rads_per_pix: float,
                                         windowtype: WindowMappingType,
                                         far_field_harvey_coeffs: HarveyScatteringCoeffs) -> Tuple[np.ndarray, Coords]:
        """
        Fits a modified harvey point spread function to the measured VIS psf.
        This currently only fits in the horizontal direction and is manually adjusted
        to fit to the regions of interest
        :return:
        """

        center = Coords(y=Ny // 2 - 0.5, x=Nx // 2 - 0.5, windowtype=windowtype)
        psf = np.zeros([Ny, Nx])
        r = PSFTools.get_radial_distance_of_pixels(psf, center)
        angle = r * rads_per_pix
        psf = cls.modified_harvey_stray_light_bsdf(angle, far_field_harvey_coeffs)
        return psf, center
