import numpy as np
from .. configuration import OpticsSpecs, AltiusConfigurationKey, PSFSpecs, Coords, WindowMappingType
from .. configuration import HarveyScatteringCoeffs
from . import PSFTools, ModifiedHarveyPSF


# ------------------------------------------------------------------------------
#           class NOOFStrayPSF
# ------------------------------------------------------------------------------
class NOOFStrayPSF:
    # ------------------------------------------------------------------------------
    #           __init__
    # ------------------------------------------------------------------------------
    def __init__(self, config: AltiusConfigurationKey):

        self._config = config
        self._optics = OpticsSpecs(config)
        self._psfspecs = PSFSpecs(self._config)[config.wavelength]
        self._harvey_coeffs: HarveyScatteringCoeffs = self._psfspecs.far_field_harvey_coeffs
        self._noof_larger_than_detector: bool = self._optics.detector_specs.window_hierarchy.NOOF_larger_than_detector
        self._rawpsf: np.ndarray = None                                                                                 # Cached PSF from file. This is the size of the detector window
        self._raw_psf_center: Coords = None                                                                             # center coordinates of the PSF

    # ------------------------------------------------------------------------------
    #           _load_internal_psf
    # ------------------------------------------------------------------------------
    def _load_internal_psf(self):
        if (self._rawpsf is None):
            apsf, self._raw_psf_center = PSFTools.load_psf(self._config)                                                # Load in the detector window from file
            Dy, Dx = apsf.shape                                                                                         # get the shape of the image
            assert (Dy == self._optics.detector_specs.window_specs.Dy) and (Dx == self._optics.detector_specs.window_specs.Dx)  # and verify it is the size of the detector window
            self._rawpsf = PSFTools.psf_knock_out_primary_lobe(apsf)                                                    # Then knock out the central lobe as that is not needed for stray PSF
        return self._rawpsf, self._raw_psf_center

    # ------------------------------------------------------------------------------
    #           fullconv_image
    # ------------------------------------------------------------------------------
    def fullconv_image(self) -> np.ndarray:
        """
        Generate a NOOF PSF that covers the entire FULLCONV window region. For VIS and NIR it must use a modified harvey function
        while for UV it can use the measured detector window array.

        Returns
        -------
        np.ndarray[Ny, Nx]
            A 2D image the size of the FULLCONV window
        """
        if self._noof_larger_than_detector:
            psf_fullconv, full_center = ModifiedHarveyPSF.fullconv_image_from_modified_harvey_curve(self._optics, self._harvey_coeffs)      # This is PSF that covers the entire area using a modified HArvey function
        else:                                                                                                                               # otherwise the full convolution window is same size as detector window
            detectorpsf, full_center = self._load_internal_psf()                                                                                    # so we can use the measured PSF
            psf_fullconv = self._optics.detector_specs.window_transforms.detector_to_fullconv(detectorpsf, centered_offsets=True)
        return psf_fullconv

    # ------------------------------------------------------------------------------
    #           detector_image
    # ------------------------------------------------------------------------------
    def detector_image(self) -> np.ndarray:
        """
        Returns a stray light PSF the size of the detector window. This will always use a PSF image. The primary central lobe
        is always knocked out so the image only represents stray light.

        Returns
        -------
        np.ndarray[Ny, Nx]
            A 2D image the size of the detector window
        """

        if self._noof_larger_than_detector:
            Dy = self._optics.detector_specs.window_specs.Dy
            Dx = self._optics.detector_specs.window_specs.Dx
            detectorpsf, full_center = ModifiedHarveyPSF.image_from_modified_harvey_curve(Dy, Dx, self._optics.radians_per_detector_pixel, WindowMappingType.WINDOW_DETECTOR, self._harvey_coeffs)
        else:                                                                                                                       # otherwise the full convolution window is same size as detector window
            detectorpsf, full_center = self._load_internal_psf()                                                                    # so we can use the measured PSF
        return detectorpsf
