from typing import Tuple, Union, Dict
import os.path
from datetime import datetime
import numpy as np
import scipy.io
import scipy.optimize
import scipy.interpolate
from l1c_deconvolution_matrix import load_from_netcdf
from ..configuration import PSFSpecs, AltiusWindowSpecs, Coords, AltiusConfigurationKey
from ..configuration.window_mapping import altius_window_specs_from_dict
from .psf_file_io import _write_psf_to_netcdf, _load_psf_from_netcdf, _load_matlab_format
from .convolution_matrix import L1CConvolutionMatrix


# ------------------------------------------------------------------------------
#           PSFTools
# ------------------------------------------------------------------------------
class PSFTools:
    """
    A class used to provide several utilities for processing and analysizing PSF arrays.
    """
    def __init__(self):
        pass

    # ------------------------------------------------------------------------------
    #           peak_from_parabola_to_max
    # ------------------------------------------------------------------------------
    @classmethod
    def _peak_from_parabola_to_max(cls, x: np.ndarray, y: np.ndarray, maxindex=None) -> float:
        """
        Fits a parabola to the sharp peak in the one dimensional data given by X and Y. ALgorithm finds the maximum value
        and steps down both sides of the curve until it is below half height. This guarantees at least 3 points. It then fits
        a parabola to the data points and finds the peak location from the maximum value of the parabola.

        Parameters
        ----------
        x: np.ndarray (N)
            The array of abscissa values. For pixels we recommned that 0.5 represent the center of a pixel, 0.0 is the
            left/lower edge and 1.0 is the right/upper edge.
        y: np.ndarray(n)
            The array of ordinate values. It is is assumed this is sharply peaked at one location
        maxindex: int or None
            The index of the maximum value. If None the maximum index is calculated.

        Returns
        -------
        Tuple[float, float]
            The x,y location of the peak.
        """

        if (maxindex is None):
            maxindex = np.argmax(y)
        maxval = y[maxindex]
        xlower = maxindex - 1
        while (y[xlower] > maxval * 0.5):
            xlower -= 1
        xupper = maxindex + 1
        while (y[xupper] > maxval * 0.5):
            xupper += 1
        xupper += 1
        index = range(xlower, xupper)
        coeffs = np.polyfit(x[index], y[index], 2)
        b = coeffs[1]
        a = coeffs[0]
        pk = -b / (2.0 * a)
        return pk

    # ------------------------------------------------------------------------------
    #           def fit_psf_peak_location(psf: np.ndarray) -> Tuple[float, float]:
    # ------------------------------------------------------------------------------
    @classmethod
    def fit_psf_peak_location(cls, psf: np.ndarray) -> Tuple[float, float]:
        """
        Finds the central peak location of a PSF by fitting parabolas to the peak location along the peak row and
        peak column. The Center coords are defined such that 0.0 corresponds to the left/lower edge of the first pixel,
        0.5 is the center and 1.0 is the right/upper edge.

        Parameters
        ----------
        psf: np.ndarray[Ny, Nx]
            the 2d image of the PSF.

        Returns
        -------
        Tuple[ float, float]
            A two element tuple. The first element is Y coordinate of the peak. The second element is the X coordinate of
            the peak.

        """
        index = np.argmax(psf)
        center_locn = np.unravel_index(index, psf.shape)
        nominal_y0 = center_locn[0]
        nominal_x0 = center_locn[1]
        x = np.arange(psf.shape[1]) + 0.5
        y = np.arange(psf.shape[0]) + 0.5
        row = psf[nominal_y0, :]
        col = psf[:, nominal_x0]
        xpeak = cls._peak_from_parabola_to_max(x, row, maxindex=nominal_x0)
        ypeak = cls._peak_from_parabola_to_max(y, col, maxindex=nominal_y0)
        return xpeak, ypeak

    # ------------------------------------------------------------------------------
    #           get_radial_distance_of_pixel
    # ------------------------------------------------------------------------------
    @classmethod
    def get_radial_distance_of_pixels(cls, psf: np.ndarray, center_coords: Coords) -> np.ndarray:
        """
        Gets the radial distance of each pixel in a PSF from a given center

        Parameters
        ----------
        psf: np.ndarray[Ny, Nx]
            the 2d image of the PSF.

        center_coords: Coords
            The coordinates of the center. The center coordinates are given so that the center of a pixel is on the 0.5 value.

        Returns
        -------
        np.ndarray[Ny,Nx]
            An array that contains the radial distance of each pixel from the center.
            the peak.

        """

        x0 = center_coords.x
        y0 = center_coords.y
        Ny = psf.shape[0]
        Nx = psf.shape[1]
        xpos2 = np.square((x0 - 0.5) - np.arange(Nx)).reshape([1, Nx])
        ypos2 = np.square((y0 - 0.5) - np.arange(Ny)).reshape([Ny, 1])
        radius1 = np.tile(xpos2, [Ny, 1])
        radius2 = np.tile(ypos2, [1, Nx])
        radius = np.sqrt(radius1 + radius2)
        return radius

    # ------------------------------------------------------------------------------
    #           get_dx_and_dy_distances_of_pixels
    # ------------------------------------------------------------------------------
    @classmethod
    def get_dx_and_dy_distances_of_pixels(cls, psf: np.ndarray, center_coords) -> Tuple[np.ndarray, np.ndarray]:
        """
        Get the :math:`\Delta X` and :math:`\Delta Y` distances of each pixel from a given center.

        Parameters
        ----------
        psf: np.ndarray[Ny, Nx]
            the 2d image of the PSF.

        center_coords: Coords
            The coordinates of the center. The center coordinates are given so that the center of a pixel is on the 0.5 value.

        Returns
        -------
        Tuple[np.ndarray[Ny,Nx], np.ndarray[Ny,Nx]]
            A two element tuple. The first element is the array of :math:`\Delta X`. The second elements is the array of :math:`\Delta Y`
        """
        x0 = center_coords[0]
        y0 = center_coords[1]
        Ny = psf.shape[0]
        Nx = psf.shape[1]
        xpos = ((x0 - 0.5) - np.arange(Nx)).reshape([1, Nx])
        ypos = ((y0 - 0.5) - np.arange(Ny)).reshape([Ny, 1])
        dx = np.tile(xpos, [Ny, 1])
        dy = np.tile(ypos, [1, Nx])
        return dx, dy

    # ------------------------------------------------------------------------------
    #           apply_vis_psf_vignetting_curve
    # ------------------------------------------------------------------------------
    @classmethod
    def apply_vis_psf_vignetting_curve(cls, signal: np.ndarray, angles_vertical: np.ndarray) -> np.ndarray:
        """
        Preliminary version that uses hard coded vignetting. Apply a vignetting curve to a signal.

        Parameters
        ----------
        signal: np.ndarray[Ny, Nx]
            the input 2d image. Vignetting will be applied to this image.

        angles_vertical: np.ndarray[Ny]
            The angle from the boresight of the vertical coordinates of the center. The hard-coded vignetting table is interpolated
            to thes evalues.

        Returns
        -------
        np.ndarray[Ny,Nx]
            A new image with vignetting applied.
        """

        vrtangle_table = np.array((-1.0, -0.0785, -0.0495, -0.0175, 1))
        vignette_table = np.array((0.0, 1.0, 1.0, 1.0, 1.0))
        v = np.interp(angles_vertical, vrtangle_table, vignette_table)
        nh = signal.shape[1]
        newsignal = np.array(signal)
        for i in range(nh):
            newsignal[:, i] *= v
        return newsignal

    # ------------------------------------------------------------------------------
    #           shift_central_area
    # ------------------------------------------------------------------------------
    @classmethod
    def _shift_central_area(cls, psf: np.ndarray, y0: float, x0: float, deltay: float, deltax: float, L: int=6):
        """
        Spatially shifts the central area of a PSF. Its intended only for small movements (eg 0.5 pixels). This uses
        high resolution sampling to move the central area.

        Parameters
        ----------
        psf: np.ndarray[Ny, Nx]
            the input 2d image. Vignetting will be applied to this image.

        y0: float
            The nominal y coordinate of the center of the PSF.

        x0: float
            The nominal x corodinate of the center of the PSF.

        deltay: float
            The amount in pixels to shift the image in y.

        deltay: float
            The amount in pixels to shift the image in x.

        L: int
            The size of the region in which to consider shifting. Pixels outside this region are not shifted. Default = 6

        Returns
        -------
        Tuple[np.ndarray[Ny,Nx], np.ndarray[Ny,Nx]]
            Two element tuple. The first element is shifted PSF image. The second element is the base image of the original PSF
        """

        samps_per_pix = 100
        N = L * samps_per_pix
        iy0 = int(y0) - L // 2
        iy1 = iy0 + L
        ix0 = int(x0) - L // 2
        ix1 = ix0 + L
        data = psf[iy0:iy1, ix0:ix1]
        base = np.min(data)
        data = data - base
        highres = np.zeros([N, N])

        for iy in range(L):
            for ix in range(L):
                ly0 = samps_per_pix * iy
                lx0 = samps_per_pix * ix
                highres[ly0:ly0 + samps_per_pix, lx0:lx0 + samps_per_pix] = data[iy, ix]

        yshift = int(deltay * samps_per_pix + 0.5)
        xshift = int(deltax * samps_per_pix + 0.5)
        highshift = np.roll(highres, (yshift, xshift), axis=(0, 1))
        lowshift = np.sum(highshift.reshape([L, samps_per_pix, L, samps_per_pix]), axis=(1, 3)) / (samps_per_pix * samps_per_pix)
        lowshift = lowshift + base
        return lowshift, base

    # ------------------------------------------------------------------------------
    #           shift_psf_to_image_centre
    # ------------------------------------------------------------------------------
    @classmethod
    def shift_psf_to_image_centre(cls, psf: np.ndarray, y0: float, x0: float) -> Tuple[np.ndarray, float, float]:
        """
        Shifts the PSF so the center of the PSF is properly centered for the FFT convolution. This means placing
        the center of the PSF at Y = 999.5 and X at 751.5 for the detector window

        Most of the image is shifted with scipy.interpolate.interpn but special care is taken with the narrow central peak.
        The pixels in the vicinity of the narrow central peak are subdivided into 1/100 pixel areas. These pixels are then
        shifted to the origin (typically 0.5 pixels)

        Parameters
        ----------
        psf: np.ndarray[Ny, Nx]
            The psf image

        y0: float
            The Y coordinate of the PSF center.

        x0: float
            The X coordinate of the PSF center.

        Returns
        -------
        Tuple[ psf, y0, x0]
            Three element tuple. The first element is the shifted PSF. The second element is the new Y center of the image.
            The third element is the new X center of the psf image.
        """

        Dy, Dx = psf.shape
        logpsf = np.log(psf)
        bad = logpsf < -30.0
        logpsf[bad] = -30.0
        newx0 = float(Dx // 2 - 1) + 0.5
        newy0 = float(Dy // 2 - 1) + 0.5
        deltax = (newx0 - x0)
        deltay = (newy0 - y0)

        assert((np.abs(deltay) < 1.1) and (np.abs(deltax) < 1.1))                         # We can only do small shifts for the moment. Update the code for bigger shifts
        L = 6                                                                               # Look at a 6 x 6 pixel region around the center.
        lowshift, base = cls._shift_central_area(psf, y0, x0, deltay, deltax, L=L)         # and shift the central region
        iy0 = int(y0) - L // 2
        iy1 = iy0 + L
        ix0 = int(x0) - L // 2
        ix1 = ix0 + L

        x = np.arange(Dx)
        y = np.arange(Dy)
        actualx = x - deltax
        actualy = y - deltay
        ax = np.tile(actualx.reshape([1, Dx]), [Dy, 1])
        ay = np.tile(actualy.reshape([Dy, 1]), [1, Dx])
        locations = np.zeros([Dy, Dx, 2])
        locations[:, :, 1] = ax
        locations[:, :, 0] = ay
        points = (y, x)
        newpsf = scipy.interpolate.interpn(points, psf, locations, bounds_error=False, fill_value=1.0E-15)
        newpsf[iy0:iy1, ix0:ix1] = lowshift
        return newpsf, newy0, newx0

    # ------------------------------------------------------------------------------
    #           load_psf_from_filename
    # ------------------------------------------------------------------------------
    @classmethod
    def load_psf_from_filename(cls, filename: str) -> np.ndarray:
        """
        Loads in a PSF array from a file. It currently only supports loading in MATLAB files

        Parameters
        ----------
        filename: str
            The full name of the file to be loaded in. If the file extension is '.mat' it is assumed to be a Matlab SAVE file
            with the PSF stored in a variable called ``psf_uv``, ``psf_vis`` or ``psf_nir``. Other files are assumed to be netcdf

        Returns
        -------
        np.ndarray(Ny, Nx)
            A 2-D array storing the PSF array. It is usually the same size as the detector window
        """

        root, ext = os.path.splitext(filename)
        ext = ext.lower()
        if (ext == '.mat'):
            psf = _load_matlab_format(filename)
        else:
            raise "Netcdf PSF reader is not yet implemented for file sof type {}".format(ext)
        return psf

    # ------------------------------------------------------------------------------
    #           load_convolution_matrix
    # ------------------------------------------------------------------------------
    @classmethod
    def load_psf(cls,
                 config: AltiusConfigurationKey,
                 psf_image_index: int = 0) -> Tuple[np.ndarray, Coords]:
        """
        Loads the convolution matrix from a netcdf file. It is assumed this convolution matrix was written using
        method ``write_to_netcdf``

        Parameters
        ----------
        filename: str
            The name of the netcdf file. It must exist, be accessible and be a netcdf file created by method ``write_to_netcdf``.

        Returns
        -------
        Tuple[ convolution_matrix, window_specs, psf_location_array, approximation ]
            Returns a three element tuple. The first element is the convolution matrix stored in the netcdf file. It is returned
            as a numpy array( Ny,Nx). The 2nd element is a dictionary listing the window specifications used to create the convolution
            matrix. The third element is a  numpy array of dimensions (3, Numpsf) that specify the (Y,X,) location of the central lobe of
            each PSF and the window type of the coordinate (2 = NOOF window, 3 = detector window, 4 = Science window).
        """
        psf_settings = PSFSpecs(config)[config.wavelength]                                                         # Fetch the PSF wavelength settings for the requested channel and date
        entry = psf_settings[psf_image_index]
        filename = entry.psf_filename
        center = entry.center_of_psf
        psf = cls.load_psf_from_filename(filename)
        return psf, center

    # ------------------------------------------------------------------------------
    #           load_convolution_matrix
    # ------------------------------------------------------------------------------
    @classmethod
    def psf_knock_out_primary_lobe(cls, psf: np.ndarray, center: Coords = None, num_central_pixels: int = 3) -> np.ndarray:
        """
        Loads the convolution matrix from a netcdf file. It is assumed this convolution matrix was written using
        method ``write_to_netcdf``

        Parameters
        ----------
        filename: str
            The name of the netcdf file. It must exist, be accessible and be a netcdf file created by method ``write_to_netcdf``.

        Returns
        -------
        Tuple[ convolution_matrix, window_specs, psf_location_array, approximation ]
            Returns a three element tuple. The first element is the convolution matrix stored in the netcdf file. It is returned
            as a numpy array( Ny,Nx). The 2nd element is a dictionary listing the window specifications used to create the convolution
            matrix. The third element is a  numpy array of dimensions (3, Numpsf) that specify the (Y,X,) location of the central lobe of
            each PSF and the window type of the coordinate (2 = NOOF window, 3 = detector window, 4 = Science window).
        """
        if (center is not None):
            x0 = int(center.x - 0.5)
            y0 = int(center.y - 0.5)
        else:
            x0 = psf.shape[1] // 2 - 1
            y0 = psf.shape[0] // 2 - 1

        newpsf = np.array(psf)
        n1 = num_central_pixels + 1
        n = num_central_pixels
        v = 0.25 * (newpsf[y0 - n1, x0] + newpsf[y0 + n1, x0] + newpsf[y0, x0 - n1] + newpsf[y0, x0 + n1])
        newpsf[y0 - n: y0 + n, x0 - n: x0 + n] = v
        return newpsf

    # ------------------------------------------------------------------------------
    #           load_convolution_matrix
    # ------------------------------------------------------------------------------
    @classmethod
    def load_convolution_matrix(cls, config: AltiusConfigurationKey,
                                matrix_type='infield',
                                detector_window_id='standard') -> L1CConvolutionMatrix:
        """
        Loads the convolution matrix from a netcdf file. It is assumed this convolution matrix was written using
        method ``write_to_netcdf``

        Parameters
        ----------
        filename: str
            The name of the netcdf file. It must exist, be accessible and be a netcdf file created by method ``write_to_netcdf``.

        Returns
        -------
        Tuple[ convolution_matrix, window_specs, psf_location_array, approximation ]
            Returns a three element tuple. The first element is the convolution matrix stored in the netcdf file. It is returned
            as a numpy array( Ny,Nx). The 2nd element is a dictionary listing the window specifications used to create the convolution
            matrix. The third element is a  numpy array of dimensions (3, Numpsf) that specify the (Y,X,) location of the central lobe of
            each PSF and the window type of the coordinate (2 = NOOF window, 3 = detector window, 4 = Science window).
        """
        wavelength_settings = PSFSpecs(config)[config.wavelength]                                                         # Fetch the PSF wavelength settings for the requested channel and date
        convolution_settings = wavelength_settings.convolution_matrix_settings                                              # Get the convolution settings from the wavelength entry
        convolution_settings.set_detector_window(detector_window_id)                                                        # Get the convolution matrices for the given detector windowing configuration`
        if matrix_type == 'infield':
            filename = convolution_settings.infield_matrix_settings['filename']
        elif matrix_type == 'horizontal_noof':
            filename = convolution_settings.horizontal_noof_matrix_settings['filename']
        else:
            raise ValueError('Unrecognized matrix type<{}>. Only <infield> and <horizontal_noof> are allowed'.format(matrix_type))
        matrix, windowspecsdict, locn = load_from_netcdf(filename)
        specs = altius_window_specs_from_dict(windowspecsdict)
        return L1CConvolutionMatrix(matrix, specs)
