from typing import Tuple
import os.path
import numpy as np
import matplotlib.pyplot as plt
from .load_legacy_psf_matlab_file import load_legacy_psf_matlab_file


# ------------------------------------------------------------------------------
#           peak_from_parabola_to_max
# ------------------------------------------------------------------------------
def _peak_from_parabola_to_max(x, y, maxindex, fignum=20):

    maxval = y[maxindex]
    xlower = maxindex - 1
    while (y[xlower] > maxval * 0.5):
        xlower -= 1
    xupper = maxindex + 1
    while (y[xupper] > maxval * 0.5):
        xupper += 1
    xupper += 1
    index = range(xlower, xupper)
    coeffs = np.polyfit(x[index], y[index], 2)
    b = coeffs[1]
    a = coeffs[0]
    pk = -b / (2.0 * a)

    xfine = np.linspace(x[index[0]], x[index[-1]], 50)
    yfit = np.polyval(coeffs, xfine)
    bigrange = range(maxindex - 15, maxindex + 15)
    plt.figure(fignum)
    plt.clf()
    plt.plot(x[bigrange], y[bigrange], 'ko-')
    plt.plot(xfine, yfit, 'r-')
    plt.plot([pk, pk], [0, maxval], 's-')
    pk = int(1000 * pk + 0.5) / 1000.0
    return pk


# ------------------------------------------------------------------------------
#           find_central_psf_peak_location
# ------------------------------------------------------------------------------
def find_central_psf_peak_location(psf_matlab_filename: str, fignum=None) -> Tuple[float, float]:
    """
    Finds the central peak location of a PSF supplied in a matlab file. It produces 3 plots that demonstrate the
    fit quality and prints the center determined. The Center coords are defined such that 0.0 corresponds to the
    left/lower edge of the first pixel, 0.5 is the center and 1.0 is the right/upper edge.

    Parameters
    ----------
    psf_matlab_filename: str
        The matlab filename containing the PSF

    Returns
    -------
    Tuple[ float, float]
        A two element tuple. The first element is Y coordinate of the peak. The second element is the X coordinate of
        the peak.

    """

    psf = load_legacy_psf_matlab_file(psf_matlab_filename)
    index = np.argmax(psf)
    center_locn = np.unravel_index(index, psf.shape)
    nominal_y0 = center_locn[0]
    nominal_x0 = center_locn[1]
    x = np.arange(psf.shape[1]) + 0.5
    y = np.arange(psf.shape[0]) + 0.5

    if (fignum is not None):
        row = psf[nominal_y0, :]
        col = psf[:, nominal_x0]
        basedir, name = os.path.split(psf_matlab_filename)
        plt.figure(1 + fignum)
        plt.clf()
        plt.imshow(np.log10(psf), origin='lower', aspect='auto')
        plt.colorbar()
        plt.xlabel('X pixels')
        plt.ylabel('Y pixels')
        plt.title('Log10 PSF from {}'.format(name))

        xpeak = _peak_from_parabola_to_max(x, row, nominal_x0, fignum=2 + fignum)
        plt.xlabel('Image X axis, pixels')
        plt.ylabel('Signal')
        plt.title('Horizontal row through PSF max. X Peak at {:7.3f} pixels'.format(xpeak))

        ypeak = _peak_from_parabola_to_max(y, col, nominal_y0, fignum=3 + fignum)
        plt.xlabel('Image Y axis, pixels')
        plt.ylabel('Signal')
        plt.title('Vertical column through PSF max. Y peak at {:7.3f} pixels'.format(ypeak))

        print('PSF central peak of {} at [{:7.3f}, {:7.3f}]'.format(name, ypeak, xpeak))
    return ypeak, xpeak


if __name__ == '__main__':
    find_central_psf_peak_location(r'C:\Users\nickl\Documents\Work\software\ARG_Instruments\QinetiQ\ancillary_data\PSF\2020-10-20-PSF\PSF\psf_uv_300nm_EOL.mat', fignum=0)
    find_central_psf_peak_location(r'C:\Users\nickl\Documents\Work\software\ARG_Instruments\QinetiQ\ancillary_data\PSF\2020-10-20-PSF\PSF\psf_uv_350nm_EOL.mat', fignum=10)
    find_central_psf_peak_location(r'C:\Users\nickl\Documents\Work\software\ARG_Instruments\QinetiQ\ancillary_data\PSF\2020-10-20-PSF\PSF\psf_vis_550nm_EOL.mat', fignum=20)
    print('Done')