import math
import numpy as np
import sasktran as sk
from skplatform import Platform
from skimpy import FrontEndOptics, SpectralWindow, FOV_Rectangle, WavelengthGaussianSRF, FrontEndRadianceGeneratorHR
from altius_l1pp_preflight.configuration.load_specs import load_specs


# ------------------------------------------------------------------------------
#           make_altius_frontend
# ------------------------------------------------------------------------------
def make_altius_wide_angle_frontend(wavelen: float):

    front_end_optics = FrontEndOptics()
    # specs = ALTIUS_INSTRUMENT_SPECS_VIS('raw')

    radiansperpix = 2.32258E-05
    upperoffset = -0.08
    maxfov = 0.08 - upperoffset
    npix = int(maxfov / radiansperpix + 0.5)
    angle_vertical = np.arange(npix) * radiansperpix + upperoffset
    angle_horiz = (np.arange(1504) - 751) * radiansperpix
    wav0 = WavelengthGaussianSRF(1, wavelen, 1.3)                                                                    # wavelength region 1. 30 wavelengths across 450 nanometers, FWHM = 0.53 nm

    for i in range(npix):
        fov = FOV_Rectangle(Hrange=(1, 0, radiansperpix),                                                               # Horizontal field of view is 5 points spanning 0.0 to 0.5 degrees in front_end_optics control frame
                            Vrange=(1, -angle_vertical[i], -(angle_vertical[i] + radiansperpix)))                       # Vertical field of view is 3 points spanning 0.0 to 20 arc seconds in front_end_optics control frame
        front_end_optics.add_spectral_window(SpectralWindow(front_end_optics, fov, wav0))                              # define the spectral_windows used in the simulation
    return front_end_optics, angle_vertical, angle_horiz


# ------------------------------------------------------------------------------
#           def make_altius_platform():
# ------------------------------------------------------------------------------
def make_altius_platform():

    platform = Platform()
    utc = ['2020-07-15 14:00:00.0000000']
    observer = [40.0, 0.0, 50000.0, 0.0, 691000.0]
    tanpoint = [50000.0, 45.0]
    platform.add_measurement_set(utc, ('looking_at_llh', observer), ('tangent_altitude', 'limb', tanpoint))
    return platform


# ------------------------------------------------------------------------------
#           make_atmosphere
# ------------------------------------------------------------------------------
def make_altius_atmosphere(albedo: float) -> sk.Atmosphere:

    atmosphere = sk.Atmosphere()
    atmosphere.brdf = albedo
    atmosphere['air'] = sk.Species(sk.Rayleigh(), sk.MSIS90())
    atmosphere['ozone'] = sk.Species(sk.O3DBM(), sk.Labow())
    return atmosphere


# ------------------------------------------------------------------------------
#           make_altius_generator
# ------------------------------------------------------------------------------
def make_altius_generator():
    sasktran_options = {'top_of_atmosphere_altitude': 110000.0}
    generator = FrontEndRadianceGeneratorHR(sasktran_options=sasktran_options)
    return generator


# ------------------------------------------------------------------------------
#           make_altius_signals
# ------------------------------------------------------------------------------
def make_altius_signal_vs_angle(wavelen: float, albedo: float):

    platform = make_altius_platform()
    frontend, anglevert, anglehoriz = make_altius_wide_angle_frontend(wavelen)
    atmosphere = make_altius_atmosphere(albedo)
    generator = make_altius_generator()
    values = generator.calculate_radiance(frontend, platform, atmosphere, toa_albedo_to_radiance_method='preconvolved_solar')
    npts = len(values)
    signal = np.zeros([npts])
    for i, entry in enumerate(values):
        signal[i] = entry.radiance[0, 0, 0]
    return np.flip(-anglevert), anglehoriz, np.flip(signal)


# ------------------------------------------------------------------------------
#           Class AltiusPlatformSimple
# ------------------------------------------------------------------------------

class AltiusPlatformSimple:

    def __init__(self):

        config = load_specs('simple_orbit_params.yaml')
        self._nominal_altitude = config['nominal_altitude_km']
        self._local_time_ascending_node = config['local_time_ascending_node']
        self._boresight_tangent_altitude = config['boresight_tangent_altitude']

    @property
    def Re(self):
        return 6371000.0
    # ------------------------------------------------------------------------------
    #           angle_to_height
    # ------------------------------------------------------------------------------

    def angle_to_height(self, angle_radians, boresight_altitude_m: float = None, max_altitude: float = 100000.0, min_altitude: float = -10000.0):
        """
        Converts a set of vertical angles from the boresight into tangent heights in meters using simple spherical geometry. The
        angle is zero at the boresight (approx center) of the detector and it is oconfigured Angle is positive upwards.

        Parameters
        ----------
        angle_radians : np.ndarray[N]
            An array of N angles measured in radians from the boresight. Angles are positive upwards and are zero at the boresight (near the middle of the detector)
        boresight_altitude_m: float
            The altitude of the boresight in meters above sea-level. If not supplied a default value is used from the configuration file (typically 50 km).
        max_altitude: float
            All altitudes value above this upper limit are returned as nan. Default is 100 km
        min_altitude: float
            All altitudes below this lower limit are returned as nan. Default is -10 km

        Returns
        -------
        np.ndarray[N]
            Returns the nominal tangent height in meters of each angle as an array on N elements. Values above and below the heighjt range limits are returned as nan.
        """

        if boresight_altitude_m is None:
            boresight_altitude_m = self._boresight_tangent_altitude

        Rs = self.Re + self._nominal_altitude
        Re = self.Re
        theta0 = math.acos((Re + boresight_altitude_m) / Rs)
        theta = theta0 - angle_radians
        h = Rs * np.cos(theta) - Re
        bad = (h > max_altitude) | (h < min_altitude)
        h[bad] = np.nan
        return h
