from typing import Union, Sequence
import math
import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
import os.path
import scipy.io
import scipy.interpolate
import scipy.signal
from altius_l1pp_preflight.configuration import OpticsSpecs, DetectorSpecs, Coords, WindowMappingType

# from altius_l1pp.altius_deconvolution import AltiusViewportDefn, AltiusViewport


# ------------------------------------------------------------------------------
#           AltiusSimulator_FrontEndRadiance
# ------------------------------------------------------------------------------

class AltiusSimulator_FrontEndRadiance:

    def __init__(self, optics: OpticsSpecs):

        self._optics = optics
        self._current_full_image = None
        self._current_full_error = None

    # ------------------------------------------------------------------------------
    #           add_read_out_noise
    # ------------------------------------------------------------------------------
    def _add_read_out_and_shot_noise(self):
        self._current_full_error += np.sqrt(self._current_full_image * self._electrons_per_dn + self._read_out_noise * self._read_out_noise) / self._electrons_per_dn

    # ------------------------------------------------------------------------------
    #           set_new_current_image
    # ------------------------------------------------------------------------------
    def _set_new_current_image(self, fullimage: np.ndarray):

        self._current_full_image = np.array(fullimage)
        self._current_full_error = np.zeros_like(fullimage)
        self._add_read_out_and_shot_noise()

    # ------------------------------------------------------------------------------
    #           signal_with_random_noise
    # ------------------------------------------------------------------------------
    def signal_with_random_noise(self):

        noisey_signal = np.random.default_rng().normal(self._current_full_image, self._current_full_error)
        return noisey_signal

    # ------------------------------------------------------------------------------
    #           load_front_end_radiance
    # ------------------------------------------------------------------------------

    def load_front_end_radiance(self,
                                wavelen_nm: float,
                                signal_viewport: Union[AltiusViewportDefn, AltiusViewport]=None,
                                maxsignal_in_dn: float = 14000.0,
                                make_plots: bool=False):
        """
        Loads in a front end radiance

        Parameters
        ----------
        fov : float
            The field of view of the science region in Radians. Default is 0.034 Radians
        """
        if signal_viewport is None:
            signal_viewport = self._optics.viewport_definition

        if isinstance(signal_viewport, AltiusViewport):
            signal_viewport = signal_viewport.viewport_definition

        rawimage_nx = signal_viewport.rawimage_nx
        rawimage_ny = signal_viewport.rawimage_ny
        viewport_y0 = signal_viewport.viewport_y0                                                                       # Y coordinate in raw detector pixels of bottom left corner of science viewport_definition
        viewport_ny = signal_viewport.viewport_ny                                                                       # Y height of science viewport_definition in raw detector pixels.
        viewport_y1 = viewport_y0 + viewport_ny                                                                         # Y coordinate in raw detector pixels of top right corner + 1 of science viewport_definition

        basedir = r'E:\altius_data'
        filename = os.path.join(basedir, 'ALTIUS_SPS_DATA_v.1.0.3_20200514T014136580', 'V1.0.3', 'BISA', 'PS-3_Scene inputs', 'PS-3.1-Limb_high_radiance.nc')
        datset = xr.load_dataset(filename)
        radiance = np.array(datset.radiance[:, :, 0])
        h = np.array(datset.apparent_tp_altitude)
        wavelength = np.array(datset.wavelength)

        goodheights = h >= -10.0                                                                                        # Find all the tangent heights between  -10 km and above the ground
        h = h[goodheights]                                                                                              # and only select those heights (they are actually in reverse order)
        f = scipy.interpolate.interp1d(wavelength, radiance[goodheights, :], axis=1)                                    # get the function to interpolate to the scalar wavelength of interest
        signal = f(wavelen_nm)                                                                                          # and do it.
        s = np.flip(signal)                                                                                             # Get the bright limb signals from 0 to 100 km
        h1 = np.flip(h)                                                                                                 # Get the heights in ascening geight order

        pixy = viewport_y0 + np.arange(0, viewport_ny)
        losh = self._optics.pixel_to_angle.raw_pixel_to_height(pixy) / 1000.0
        interp_sig = np.interp(losh, h1, s)                                                                             # and interpolate the height profile
        s = interp_sig.reshape([viewport_ny, 1])                                                                  # reshape in preparation for tile
        sciimage = np.tile(s, [1, rawimage_nx])                                                                  # and tile the image horizontally, same value at all locations across the whole field of view of the
        if (make_plots):
            plt.figure(1)
            plt.clf()
            plt.plot(interp_sig, losh)
            plt.xlabel('Radiance Photons/cm2/sec/ster/nm')
            plt.ylabel('Height Km')
            plt.title('Input signal for {} nm'.format(wavelen_nm))

        iy0 = viewport_y0
        iy1 = viewport_y1
        fullimage = np.zeros([rawimage_ny, rawimage_nx])                                                     # make a full detector image
        fullimage[iy0:iy1, :] = sciimage                                                                                # apply the science viewpport rowsmake the middle
        fullimage[:iy0, :] = sciimage[0, :]
        fullimage[iy1:, :] = sciimage[-1, :]
        if (make_plots):
            plt.figure(2)
            plt.clf()
            plt.imshow(fullimage)
            plt.colorbar()
            plt.xlabel('Raw detector X pixel')
            plt.ylabel('Raw detector Y pixel')
            plt.title('Near-OOF, no vignetting. {} nm'.format(wavelen_nm))

        vx = self.vignette_value_x(np.arange(rawimage_nx)).reshape([1, rawimage_nx])
        vy = self.vignette_value_y(np.arange(rawimage_ny)).reshape([rawimage_ny, 1])
        fullimage *= vx
        fullimage *= vy

        maxv = np.max(fullimage)
        factor = maxsignal_in_dn / maxv
        fullimage *= factor
        self._set_new_current_image(fullimage)
        if (make_plots):
            plt.figure(3)
            plt.clf()
            plt.imshow(fullimage)
            plt.colorbar()
            plt.xlabel('Raw detector X pixel')
            plt.ylabel('Raw detector Y pixel')
            plt.title('Near-OOF, with vignetting. {} nm'.format(wavelen_nm))

            plt.figure(4)
            plt.clf()
            plt.plot(vx[0, :], label='X vignetting')
            plt.plot(vy, label='Y vignetting')
            plt.xlabel('Raw Detector Pixels')
            plt.legend()
            plt.title('Vignetting curves')
        print("Done")
        return fullimage
