from typing import Dict, Any
import numpy as np
import sasktran as sk
import copy
from skretrieval.platforms import Platform
# from skimpy import Instrument, ForwardModel, Platform, FOV_Rectangle, Wavelength_GaussianPSF,SpectralWindow, RadianceValues_UVIS_NIR
from sasktran.species import SpeciesAerosolGloSSAC
# from ..instrument_specs import create_altius_instrument_specs, ALTIUS_INSTRUMENT_SPECS
from altius_l1pp_preflight.radiance_simulator.altius_bright_limb_signals.deprecated.simple_attitude_solution import ALTIUS_SimpleSpacecraft


# ------------------------------------------------------------------------------
#           Instrument_AltiusBrightLimb
# ------------------------------------------------------------------------------
class Instrument_AltiusBrightLimb:
    """
    This is a Skimpy Instrument used to model the bright limb signals for ALTIUS. It is meant to provide a quick way
    to get model results for typical ALTIUS conditions. It is used during the early stages of USASK work on ALTIUS
    l1pp development. By default the tangent point is located close to the equator and the satellite is in the Northern
    hemisphere looking southward toward the tangent point.

    A quick simulation can be as easy as::

        model   = Instrument_AltiusBrightLimb('vis','raw')
        results = model.calculate_signals( wavelengths=[500.0, 510.0, 520.0])

    which returns a dictionary with the wavelengths, heights, radiances, photons, toa_albedo, latitude and longitude.

    """

    def __init__(self, specs: ALTIUS_INSTRUMENT_SPECS):
        """
        Creates a skimpy instrument model for the given ALTIUS channel in bright limb mode. This model uses a simplified
        attitude solution which is good for preliminary studies.

        Parameters
        ----------
        specs : ALTIUS_INSTRUMENT_SPECS
            The specifications of the required instrument/channel. The specifications has all the details about
            field of view and pixel sizes.
        """
        super().__init__()
        self._specs: ALTIUS_INSTRUMENT_SPECS = copy.deepcopy(specs)                                    # Copy the specifications so they can't be changed outside this object
        self._simpleatt: ALTIUS_SimpleSpacecraft = None                                                      # An object that generates a simplified attitude solution
        self._lowestheight: float = 5000.0                                                  # The default height in meters of the bottom pixel
        self.utcstr: str = '2013-07-31T12:00:00.000000'                              # We use this timesamp so the Glossac aerosol database is valid
        self._wavelen: np.ndarray = None                                                      # The wavelengths used in the radiative transfer calculation
        self._waveindices: np.ndarray = None                                                      # and the indices of these wavelengths in the radiance arrays
        self._heights: np.ndarray = None                                                      # the tangent altitudes in meters of the lines of sight used in the radiative transfer
        self._fovindices: np.ndarray = None                                                      # and the indices o fthese lines of sight in the radiance arrays.
        self._lastresults: Dict[str, Any] = None                                                      # A dictionary of results from the last radiative trnsfer calculation

    # ------------------------------------------------------------------------------
    #           lowestheight
    # ------------------------------------------------------------------------------
    @property
    def lowestheight(self) -> float:
        return self._lowestheight

    # ------------------------------------------------------------------------------
    #           set_lowestheight
    # ------------------------------------------------------------------------------
    def set_lowestheight(self, h: float):
        self._lowestheight = h

    # ------------------------------------------------------------------------------
    #           Instrument_AltiusBrightLimb::specs
    # ------------------------------------------------------------------------------
    @property
    def specs(self):
        """
        Returns the instrument specs associated with this instrument
        """
        return self._specs

    # ------------------------------------------------------------------------------
    #           Instrument_AltiusBrightLimb::calculate_signals
    # ------------------------------------------------------------------------------
    def calculate_signals(self,
                          wavelengths=None,
                          wavelen_stepsize=10.0,
                          max_height_km=None,
                          albedo=0.3,
                          tanlatitude=0.0,
                          sza=40.0,
                          saa=60.0,
                          use_aerosols=True) -> Dict[str, np.ndarray]:

        self._simpleatt = ALTIUS_SimpleSpacecraft(self.specs, self._lowestheight)
        self._simpleatt.position_and_orient_satellite(tanlatitude, sza, saa)
        self._make_instrument(max_height_km=max_height_km, wavelengths=wavelengths, wavelen_stepsize=wavelen_stepsize)

        atmosphere = self._make_atmosphere(albedo=albedo, use_aerosols=use_aerosols)
        forwardmodel = self._make_forwardmodel()
        platform = self._make_platform()
        forwardmodel_options = {'solar_spectrum': 'FONTELA_UVIS_3MICRON'}
        forwardmodel.calculate_radiance(self, platform, atmosphere, forward_model_options=forwardmodel_options, geometry_options={'sun': self._simpleatt.sun})
        return self._lastresults

    # ------------------------------------------------------------------------------
    #           _make_platform
    # ------------------------------------------------------------------------------
    def _make_platform(self) -> Platform:                                        # Solar azimuth angle, horizontal azimuth of sun at tangent point measured in degree from due south +ve towards west

        platform = Platform()                                                                                           # Create a simple platform object
        platform.clear_states()                                                                                         # Clear the current observation set
        t = np.datetime64(self.utcstr)                                                                             # Set the time of the observations. This is typically a time in 2013 to make the Glossac aerosol database work
        satpos = self._simpleatt.satpos                                                                                 # Calculate the position of the spacecraft,m the look ector of the boresight and the location of the sun.
        platform.set_platform_location(xyzt=[satpos[0], satpos[1], satpos[2], t])                                   # Call Skimpy, Set the location of the satellite
        platform.set_boresight_from_limb_looking_vector(self._simpleatt.boresight)                                                         # Call Skimpy, Set the boresight to be along the look evctor
        platform.add_current_platform_state()                                                                           # Call Skimpy, Add this platform state (all the lines of sight from all the spectral windows are added).
        return platform

    # ------------------------------------------------------------------------------
    #           _make_atmosphere
    # ------------------------------------------------------------------------------
    def _make_atmosphere(self, albedo: float = 0.3,
                         use_aerosols: bool = True) -> sk.Atmosphere:

        atmosphere = sk.Atmosphere()
        atmosphere.atmospheric_state = sk.MSIS90()
        atmosphere.brdf = albedo
        atmosphere['air'] = sk.Species(sk.Rayleigh(), sk.MSIS90())
        atmosphere['ozone'] = sk.Species(sk.O3OSIRISRes(), sk.Labow())
        if (use_aerosols):
            atmosphere['aerosol'] = SpeciesAerosolGloSSAC()
        return atmosphere

    # ------------------------------------------------------------------------------
    #           ForwardModel_Spectrograph
    # ------------------------------------------------------------------------------
    def _make_forwardmodel(self) -> ForwardModel:

        model = ForwardModel('HR')
        # model.radiance_simulator().set_option_weighting_functions_for_toa_albedo( True )
        return model

    # ------------------------------------------------------------------------------
    #           _get_default_list_of_wavelengths
    # ------------------------------------------------------------------------------
    def _get_default_list_of_wavelengths(self, wavelen_stepsize):

        waverange = self.specs.wavelength_range_nm
        wavelen = np.arange(waverange[0], waverange[1], wavelen_stepsize)
        return wavelen

    # ------------------------------------------------------------------------------
    #           _make_wavelength_range
    # ------------------------------------------------------------------------------
    def _make_wavelength_range(self, wavelengths: np.ndarray = None, wavelen_stepsize=10.0, resolution=1.5) -> np.ndarray:
        """
        Returns an array of wavelengths that span the range of wavelengths used by this ALTIUS channel
        """
        self._wavelen = np.copy(wavelengths) if wavelengths is not None else self._get_default_list_of_wavelengths(wavelen_stepsize)
        Nw = self._wavelen.size
        index = []
        for iw in range(Nw):
            idx = self.add_wavelength_range(Wavelength_GaussianPSF(1, self._wavelen[iw], resolution, rtm_processing_scheme='hybrid', photon_distribution='additive'))
            index.append(idx)
        self._waveindices = np.array(index)
        return self._waveindices

    # ------------------------------------------------------------------------------
    #           _make_instrument
    # ------------------------------------------------------------------------------
    def _make_instrument(self, max_height_km, wavelengths, wavelen_stepsize):
        """
        A function to make the spectral windows
        """
        if (max_height_km is not None):
            max_height_km = (max_height_km - self._simpleatt.lowestheight / 1000.0)
        fovlist, h = self._simpleatt.make_list_of_pixel_fov(maxheightrange_km=max_height_km)
        index = []
        for entry in fovlist:                                                                                           # for each vertical pixel
            hfov = (1, entry[0][0], entry[0][1])                                                                 # define the bottom or start of the vertical pixel
            vfov = (1, entry[1][0], entry[1][1])
            fov = FOV_Rectangle(Hrange=hfov, Vrange=vfov)                                                        # Define a rectangular field of view for the science pixel fields of view for each "wavelength" spectral_window
            idx = self.add_field_of_view(fov)                                                                           # Add this field of view to our list of field of views
            index.append(idx)

        self._heights = h
        self._fovindices = np.array(index)                                                                             # Make the field of view objects each pixel in the height direction covering a height range of max_height_km
        self._waveindices = self._make_wavelength_range(wavelengths=wavelengths, wavelen_stepsize=wavelen_stepsize)   # make the array of wavelength bandpasses
        aperture_area_cm2 = self.specs.aperture_size_cm2
        numfov = self._fovindices.size
        numwav = self._waveindices.size

        for iv in range(numfov):                                                                                        # for each height
            fovidx = self._fovindices[iv]                                                                               # get the index for this field of view
            for iw in range(numwav):                                                                                    # then for all the wavelengths
                wavidx = self._waveindices[iw]                                                                          # get the index of this wavelength bandpass
                self.add_spectral_window(SpectralWindow(self, fovidx, wavidx, aperture_area_cm2))                        # create the spectral window for ALTIUS UV Wavelength 1 at this height
        return self

    # ------------------------------------------------------------------------------
    #           _get_avg_radiance
    # ------------------------------------------------------------------------------
    def _get_radiance_value(self, window: SpectralWindow) -> RadianceValues_UVIS_NIR:

        hires_radiance = window.high_resolution_radiance
        avg_radiance = RadianceValues_UVIS_NIR(window, hires_radiance.radiance, hires_radiance.toa_albedo,
                                               hires_radiance.wf,
                                               False)
        return avg_radiance

    # ------------------------------------------------------------------------------
    #           propagate_frontend_to_backend
    # ------------------------------------------------------------------------------
    def propagate_frontend_to_backend(self, platform: Platform, attitude_solution: Dict[str, np.ndarray]):
        """
        The callback function used by skimpy to allow the instrument_specs to organize the high resolution
        radiances calculed by the forward model.
        """
        numwavelen = self.wavelength_ranges.size
        numheights = self.spectral_windows.num_spectral_windows() // numwavelen
        wavelen = np.zeros([numwavelen])
        latitude = np.zeros([numheights])
        longitude = np.zeros([numheights])
        height = np.zeros([numheights])
        radiance = np.zeros([numheights, numwavelen])
        photons = np.zeros([numheights, numwavelen])
        toa = np.zeros([numheights, numwavelen])
        geo = sk.Geodetic()
        #
        for ih in range(numheights):
            for iw in range(numwavelen):
                window = self.spectral_windows[ih * numwavelen + iw]
                wavelen[iw] = window.central_wavelength()
                r = self._get_radiance_value(window)
                radiance[ih, iw] = r.radiance
                photons[ih, iw] = r.photon_flux
                toa[ih, iw] = r.toa_albedo

            look = attitude_solution['los'][:, ih, 0]
            pos = attitude_solution['pos'][:, ih, 0]
            geo.from_tangent_point(pos, look)
            latitude[ih] = geo.latitude
            longitude[ih] = geo.longitude
            height[ih] = geo.altitude

        self._lastresults = {'w': wavelen,
                             'rad': radiance,
                             'photons': photons,
                             'toa': toa,
                             'h': height,
                             'lat': latitude,
                             'lng': longitude
                             }
