from typing import List, Tuple, Any, Union
import copy
import math
import numpy as np
import sasktran as sk
from skretrieval.platforms import Platform
from skimpy import FrontEndOptics, SpectralWindow, FOV_Rectangle, WavelengthGaussianSRF, FrontEndRadianceGeneratorHR
from altius_l1pp_preflight.configuration import OpticsSpecs, AltiusConfigurationKey, OrbitParams_SimpleOrbit, WindowMappingType, Coords


# ------------------------------------------------------------------------------
#           ALTIUS_SimpleSpacecraft
# ------------------------------------------------------------------------------
class ALTIUS_SimpleSpacecraft:
    """
    This is a class that provides a simplified attitude solution for ALTIUS studies. It allows for consistent placement of
    the satellite and look vectors. It allows the sun to be placed anwywhere in the sky which is useful for many studies
    """

    def __init__(self, specs: OpticsSpecs, lowestheight=5000.0):
        """
        Creates a skimpy instrument model for the given ALTIUS channel in bright limb mode.

        """
        self._optics: OpticsSpecs = specs
        self._orbitspecs = OrbitParams_SimpleOrbit()
        self._satpos: np.ndarray = None
        self._xicf: np.ndarray = None
        self._sun: np.ndarray = None

    # ------------------------------------------------------------------------------
    #           set_lowestheight
    # ------------------------------------------------------------------------------
    def set_lowestheight(self, lowestheightkm):
        """
        Sets the lowest tangent height in kms of the lowest pixel. This will also reset the internal attitude
        solution and users must make a call to `position_and_orient_satellite` to update the internal attitude solutions.
        """
        self._satpos: np.ndarray = None
        self._xicf: np.ndarray = None               # the x unit vector in the instrument control frame. Boresight
        self._zicf: np.ndarray = None               # the z unit vector in the instrument control frame. Niominal Up, but inclined for the boresight looking at 50 km.
        self._yicf: np.ndarray = None               # the y unit vector in the instrument control frame. To the left when looking along the boresight away from the instrument.
        self._sun: np.ndarray = None

    # ------------------------------------------------------------------------------
    #           satpos
    # ------------------------------------------------------------------------------
    @property
    def satpos(self) -> np.ndarray:
        """
        Returns the current x,y,z position of the spacecraft in meters
        """
        return self._satpos

    # ------------------------------------------------------------------------------
    #           boresight
    # ------------------------------------------------------------------------------
    @property
    def boresight(self) -> np.ndarray:
        """
        Returns the current primary (central) line of sight look vector of the instrument
        as an x,y,z unit vector
        """
        return self._xicf

    # ------------------------------------------------------------------------------
    #           sun
    # ------------------------------------------------------------------------------
    @property
    def sun(self) -> np.ndarray:
        """
        Returns the current x,y,z unitvector from the Earth to the sun.
        """
        return self._sun

    # ------------------------------------------------------------------------------
    #           Instrument_AltiusBrightLimb::position_and_orient_satellite
    # ------------------------------------------------------------------------------
    def position_and_orient_satellite(self, tanpoint_latitude, sza, saa, satellite_altitude: float = None, boresight_altitude: float=None):
        """
        This positions the satellite so the tangent points are at (or close to) the given tangent point latitude.
        The satellite is always oriented so it is looking due South and the Sun is positioned so it meets the required
        SZA nad SAA at the tangent point. The boresight look vector is oriented so the boresight is
        looking at the requested boresight tangent altitude.

        Parameters
        ----------
        tanpoint_latitude: float
            The latitude of the required tangent point in degrees.  The satellite is placed so it is looking at this latitude (at zero longitude).

        sza: float
            Requested solar zenith angle. The location of the sun is placed so this SZA occurs at the tangent point.

        saa: float
            Request solar azimuth angle. The location of the sun is placed so this SAA occurs at the tangent point.

        satellite_altitude: float
            Optional. Sets the altitude of the spacecraft in meters above sea level. If not provided the value from the ``simple_orbit_params.yaml`` is used.

        boresight_altitude: float
            Optional. Sets the altitude of the boresight in meters above sea level. If not provided the value from the ``simple_orbit_params.yaml`` is used.
        """

        hsat = satellite_altitude if satellite_altitude is not None else self._orbitspecs.nominal_spacecraft_altitude
        bore_alt = boresight_altitude if boresight_altitude is not None else self._orbitspecs.boresight_tangent_altitude

        geo = sk.Geodetic()
        geo.from_lat_lon_alt(tanpoint_latitude, 0, bore_alt)                                                            # Get the vector from center of Earth to tangent point at the boresight altitude
        y = geo.local_south if tanpoint_latitude < 0.0 else -geo.local_south                                            # Look South in Southern hemisphere, Look North in Norther hemisphere
        tanpoint = geo.location                                                                                         # Get the vector from center of Earth to tangent point
        Ret = np.linalg.norm(tanpoint)                                                                                  # Get the radius of the tangent point.
        Resat = Ret + (hsat - bore_alt)                                                                                 # Get the radius of the satellite
        L = math.sqrt(Resat * Resat - Ret * Ret)                                                                        # get the approximate distance to the satellite from the tangent point
        self._satpos = tanpoint - y * L                                                                                 # and get the satellite position

        look = tanpoint - self._satpos                                                                                  # get the look vector. This will normally be due north south (the same as y)
        self._xicf = look / np.linalg.norm(look)                                                                        # but not sure what happens at the poles so I do this calculation
        geo.from_tangent_point(self._satpos, self._xicf)                                                                # and recalculate the (same) tanpoint again
        self._zicf = geo.local_up                                                                                       # get the local up as the instrument z unit vector
        self._yicf = np.cross(self._zicf, self._xicf)                                                                   # and the y unit vector from the cross product. Its to the left when looking away from the instrument along  xicf
        x = geo.local_west                                                                                              # Get the local unit vectors
        y = geo.local_south
        z = geo.local_up

        saar = math.radians(saa)                                                                                        # get the SAA in radians
        szar = math.radians(sza)                                                                                        # get the SZA in radians
        h = math.cos(saar) * y + math.sin(saar) * x                                                                     # get the horizontal unit vector of the solar unit vector
        self._sun = math.cos(szar) * z + math.sin(szar) * h                                                             # get the unit vector to the _sun.

    # ------------------------------------------------------------------------------
    #           _pixel_to_angle
    # ------------------------------------------------------------------------------
    def _pixel_to_angle(self, userpixelx: float, userpixely: float, windowtype: WindowMappingType) -> Coords:
        userpt = Coords(x=userpixelx, y=userpixely, windowtype=windowtype)                                              # Create the corodinate
        detpt = self._optics.detector_specs.window_hierarchy.convert_to(userpt, WindowMappingType.WINDOW_DETECTOR)      # map this point to the detector window
        angle = self._optics.pixel_to_radians(detpt)
        return angle

    # ------------------------------------------------------------------------------
    #           make_list_of_pixel_fov
    # ------------------------------------------------------------------------------
    def make_list_of_pixel_fov(self, windowtype=WindowMappingType.WINDOW_SCIENCE) -> Tuple[List[Any], np.ndarray]:
        """
        Calculates a list containing the angular field of view, wrt the _xicf, of each ALTIUS pixel
        in the vertical direction. Also returns an array indicating the tangent altitude of each o fthe lines of sight.
        This is a simple attitude solution: the spacecraft is at a fixed location and looking southward. It is intended
        for studies that do not require precision attitude solutions

        Returns
        -------
        A 2 elements tuple.
        * The first element is a list of pixel field of views, suitable for use by the Skimpy code. There is one entry for each pixel. Each entry
        is a two elelemnt tuple containing the horizontal and vertical field of view of each pixel that can be used by the
        Skimpy instrument.
        * The second element is a numpy array containing the tangent height of each pixel in meters.
        """

        geo = sk.Geodetic()
        userwindowmap = self._optics.detector_specs.window_hierarchy[windowtype]                                        # Get the mapping object for the requested window type
        Ny = userwindowmap.Numy()                                                                                       # Get the number of Y pixels for this window
        userx = userwindowmap.Numx() / 2.0                                                                              # Get the middle of the array for X

#        geo.from_tangent_point(self._satpos, self._xicf)
#        xprime = geo.local_west                                                                                              # Get the local unit vectors
#        y = geo.local_south
#        z = geo.local_up

        y0 = Ny / 2
        userpt = Coords(x=userx - 0.5, y=y0, windowtype=windowtype)
        leftangle = self._optics.pixel_to_radians(userpt)
        userpt = Coords(x=userx + 0.5, y=y0, windowtype=windowtype)
        rightangle = self._optics.pixel_to_radians(userpt)
        deltax = math.fabs(rightangle.x - leftangle.x)

        fov = []
        height = np.zeros([Ny])
        userpt = Coords(x=userx, y=0.0, windowtype=windowtype)                                                          # start at the bottom of the detcetor
        bottomangle = self._optics.pixel_to_radians(userpt)                                                             # Get the angle at the bottom edge of the first pixel
        for iy in range(Ny):                                                                                            # for each vertical pixel in the user window
            userpt = Coords(x=userx, y=(iy + 1), windowtype=windowtype)                                                 # get the user pixel coordinates at the top edge of the pixel
            topangle = self._optics.pixel_to_radians(userpt)                                                            # get the angle at top edge of the pixel
            theta = (topangle.y + bottomangle.y) * 0.5                                                                  # Get the midpoint of the two angles. This is the boresight o fthe pixel
            deltay = (topangle.y - bottomangle.y)                                                                       # Get the angular range across the pixel. This is the delta angle
            fov.append(((0.0, deltax), (theta, deltay)))                                                                # Define a rectangular field of view in x and y in radians. This used for the pixel's field of view in the skimpy calculations.
            los = math.cos(theta) * self._xicf + math.sin(theta) * self._zicf                                           # get the look vector from the satellite
            geo.from_tangent_point(self._satpos, los)                                                                   # and get the tangent point
            height[iy] = geo.altitude                                                                                   # append the height

        return fov, height


############################################################################################################################################
class AltiusSkimpyModel:

    def __init__(self, specs: OpticsSpecs):

        self._optics = specs
        self._orbitspecs = OrbitParams_SimpleOrbit()
        self._wavelen: np.ndarray = None
        self._platform: Platform = None
        self._atmosphere: sk.Atmosphere = None
        self._generator: FrontEndRadianceGeneratorHR = None

    # ------------------------------------------------------------------------------
    #           make_altius_frontend
    # ------------------------------------------------------------------------------
    def make_altius_wide_angle_frontend(self, wavelena: Union[List[float], float]) -> Tuple[FrontEndOptics, np.ndarray, np.ndarray]:

        front_end_optics = FrontEndOptics()

        radiansperpix = self._optics.radians_per_pixelY
        upperoffset = -0.08
        maxfov = 0.08 - upperoffset
        npix = int(maxfov / radiansperpix + 0.5)
        angle_vertical = np.arange(npix) * radiansperpix + upperoffset
        angle_horiz = (np.arange(1504) - 751) * radiansperpix
        wav0 = WavelengthGaussianSRF(1, wavelen, 1.3)                                                                    # wavelength region 1. 30 wavelengths across 450 nanometers, FWHM = 0.53 nm

        for i in range(npix):
            fov = FOV_Rectangle(Hrange=(1, 0, radiansperpix),                                                               # Horizontal field of view is 5 points spanning 0.0 to 0.5 degrees in front_end_optics control frame
                                Vrange=(1, -angle_vertical[i], -(angle_vertical[i] + radiansperpix)))                       # Vertical field of view is 3 points spanning 0.0 to 20 arc seconds in front_end_optics control frame
            front_end_optics.add_spectral_window(SpectralWindow(front_end_optics, fov, wav0))                              # define the spectral_windows used in the simulation
        return front_end_optics, angle_vertical, angle_horiz

    # ------------------------------------------------------------------------------
    #           def make_altius_platform():
    # ------------------------------------------------------------------------------
    def make_altius_platform(self) -> Platform:

        platform = Platform()
        utc = ['2020-07-15 14:00:00.0000000']
        observer = [40.0, 0.0, 50000.0, 0.0, 691000.0]
        tanpoint = [50000.0, 45.0]
        platform.add_measurement_set(utc, ('looking_at_llh', observer), ('tangent_altitude', 'limb', tanpoint))
        return platform

    # ------------------------------------------------------------------------------
    #           make_atmosphere
    # ------------------------------------------------------------------------------
    def make_altius_atmosphere(self, albedo: float) -> sk.Atmosphere:

        atmosphere = sk.Atmosphere()
        atmosphere.brdf = albedo
        atmosphere['air'] = sk.Species(sk.Rayleigh(), sk.MSIS90())
        atmosphere['ozone'] = sk.Species(sk.O3DBM(), sk.Labow())
        return atmosphere

    # ------------------------------------------------------------------------------
    #           make_altius_generator
    # ------------------------------------------------------------------------------
    def make_altius_generator(self) -> FrontEndRadianceGeneratorHR:
        sasktran_options = {'top_of_atmosphere_altitude': 110000.0}
        generator = FrontEndRadianceGeneratorHR(sasktran_options=sasktran_options)
        return generator

    # ------------------------------------------------------------------------------
    #           make_altius_signals
    # ------------------------------------------------------------------------------
    def make_altius_signal_vs_angle(self, wavelen: float, albedo: float) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:

        platform = make_altius_platform()
        frontend, anglevert, anglehoriz = make_altius_wide_angle_frontend(specs, wavelen)
        atmosphere = make_altius_atmosphere(albedo)
        generator = make_altius_generator()
        values = generator.calculate_radiance(frontend, platform, atmosphere, toa_albedo_to_radiance_method='preconvolved_solar')
        npts = len(values)
        signal = np.zeros([npts])
        for i, entry in enumerate(values):
            signal[i] = entry.radiance[0, 0, 0]
        return np.flip(-anglevert), anglehoriz, np.flip(signal)
