import numpy as np
import sasktran as sk
import math
from skimpy import Instrument, ForwardModel, Platform, FOV_Rectangle, Wavelength_GaussianPSF,SpectralWindow
from sasktran.species import SpeciesAerosolGloSSAC
from altius_l1pp_simulator.instrument_specs import create_altius_instrument_specs, ALTIUS_INSTRUMENT_SPECS

#------------------------------------------------------------------------------
#           SkimpyCommonBase
#------------------------------------------------------------------------------

class SkimpyCommonBase(Instrument):

    def __init__(self, instrumentname: str, mode: str ):
        """
        Creates a skimpy instrument model for the given ALTIUS channel. The base class includes  and given ALTIUS mode

        Parameters
        ----------
        instrumentname : str
            The name of the instrument channel.  The model will automatically select the correct detector, field of view
            and wavelength range parameters based upon this value. It must be one of 'uv', 'vis' or 'nir' only.

        mode: str
            The mode of operation. The model will automatially select the correct detector parameters etc. It must be
            either "raw" or "sci" only.
        """
        super().__init__()
        self._specs : ALTIUS_INSTRUMENT_SPECS = create_altius_instrument_specs(instrumentname, mode)
        self.centralheight = 50000.0
        self.utcstr        = '2013-07-31T12:00:00.000000'
        self._wavelen      = None
        self._waveindices  = None
        self._fovindices   = None

        #------------------------------------------------------------------------------
    #           specs
    #------------------------------------------------------------------------------

    @property
    def specs(self):
        """
        Returns the instrument specs associated with this instrument
        """
        return self._specs

    #------------------------------------------------------------------------------
    #           calculate_signals
    #------------------------------------------------------------------------------

    def calculate_signals(self, albedo = 0.3, tanlatitude= 0.0, sza = 0.0, saa = 0.0, use_aerosols= True):

        self.make_instrument( )

        centralheight = self.centralheight
        atmosphere    = self.make_atmosphere( albedo       = albedo,
                                              use_aerosols = use_aerosols )
        forwardmodel  = self.make_forwardmodel()
        platform      = self.make_platform( centralheight = centralheight,
                                            tanlatitude   = tanlatitude,
                                            sza           = sza,
                                            saa           = saa  )

        satpos, look, sun = self.calculate_satellite_look_sun( tanlatitude, sza, saa, tanpoint_height = self.centralheight)
        forwardmodel_options={'solar_spectrum' : 'FONTELA_UVIS_3MICRON'}
        forwardmodel.calculate_radiance( self, platform, atmosphere, forward_model_options=forwardmodel_options, geometry_options={'sun':sun }  )

    #------------------------------------------------------------------------------
    #           position_and_orient_satellite
    #------------------------------------------------------------------------------

    def calculate_satellite_look_sun(self, tanpoint_latitude, sza, saa, tanpoint_height  ):

        location = sk.Geodetic()
        location.from_lat_lon_alt(tanpoint_latitude, 0, 0)                                                              # Get the vector from center of Earth to tangent point at surface
        x = location.local_west                                                                                         # Get the local unit vectors
        y = location.local_south
        z = location.local_up
        Revec = location.location                                                                                       # Get the vector from center of Earth to ground
        Re    = np.linalg.norm(Revec)                                                                                   # Get the approximate value of Re. Its not exact buts is close enough for what we need
        tanpoint = Revec + z * tanpoint_height                                                                          # Get the vector to the tangent point
        Ret = np.linalg.norm(tanpoint)                                                                                  # and the length of that vector
        Resat = Re + self.specs.nominal_spacecraft_altitude_kms*1000.0
        L = math.sqrt(Resat * Resat - Ret * Ret)                                                                        # get the approximate distance to the satellite from the tangent point
        satpos = tanpoint - y * L                                                                                       # and get the satellite position
        look   = y                                                                                                      # and the look vector toward that tangent point.

        saar  = math.radians(saa)
        szar  = math.radians(sza)
        h     = math.cos( saar )*y + math.sin( saar )*x
        sun   = math.cos( szar )*z + math.sin( szar )*h

        return satpos, look, sun


    # ------------------------------------------------------------------------------
    #           ObservationSetManagerTest
    # ------------------------------------------------------------------------------

    def make_platform( self, tanlatitude   = 0.0,                                                                       # latitude of tangent point in degrees
                             sza           = 0.0,                                                                       # Solar zenith
                             saa           = 0.0  )->Platform:                                                          # Solar azimuth angle, horizontal azimuth of sun at tangent point measured in degree from due south +ve towards west


        platform = Platform()                                                                                           # Create a simple platform object
        platform.clear_states()                                                                                         # Clear the current observation set
        t        = np.datetime64(self.utcstr)                                                                           # Set the time of the observations
        satpos,look,sun = self.calculate_satellite_look_sun(  tanlatitude, sza, saa, tanpoint_height=self.centralheight) # The platform looks due North
        platform.set_platform_location( xyzt = [ satpos[0], satpos[1], satpos[2], t])                                   # Set the location of the satellite
        platform.set_boresight_from_limb_looking_vector( look )                                                         # Set the boresight to be along the look evctor
        platform.add_current_platform_state()                                                                           # Add this platform state (all the lines of sight from all the spectral windows are added).
        return platform, sun

    # ------------------------------------------------------------------------------
    #           make_spectrograph_atmosphere
    # ------------------------------------------------------------------------------

    def make_atmosphere( self,  albedo       : float = 0.3,
                                use_aerosols : bool = True )->sk.Atmosphere:

        atmosphere = sk.Atmosphere()
        atmosphere.atmospheric_state = sk.MSIS90()
        atmosphere.brdf       = albedo
        atmosphere['air']     = sk.Species(sk.Rayleigh(), sk.MSIS90())
        atmosphere['ozone']   = sk.Species( sk.O3OSIRISRes(), sk.Labow())
        if (use_aerosols) : atmosphere['aerosol'] = SpeciesAerosolGloSSAC()
        return atmosphere

    # ------------------------------------------------------------------------------
    #           ForwardModel_Spectrograph
    # ------------------------------------------------------------------------------

    def make_forwardmodel(self )->ForwardModel:

        model = ForwardModel( 'HR' )
        #model.radiance_simulator().set_option_weighting_functions_for_toa_albedo( True )
        return model

    #------------------------------------------------------------------------------
    #           _make_height_profile_of_pixel_fov
    #------------------------------------------------------------------------------

    def make_height_profile_of_pixel_fov(self, maxheightrange_km = None )->np.ndarray:
        """
        Calculates a vertical profile of Skimpy FOV_Rectangle objects, one for each pixel.
        """

        arcminsperkm    = self.specs.get_arcminutes_per_km_in_limb()
        arcminsperpix   = self.specs.arcminutes_per_pixel
        pixelsperarcmin = 1.0/arcminsperpix[0]
        Ny              = self.specs.Ny
        if (maxheightrange_km is None):
            Nv = Ny
        else:
            Nv =int((maxheightrange_km)*arcminsperkm*pixelsperarcmin +0.5)
            if (Nv > Ny): Nv = Ny
        ccdfov                    = self.specs.FOV_Radians                                                              # Get the fov of the entire detector
        bottomangle              = -ccdfov[0]/2.0                                                                       # The bottom altitude is at the bottom of the detector, halfway from the central location
        deltah                    = math.radians( arcminsperpix[1]/60.0 )                                               # The horizontal angular extent of each science pixel
        deltav                    = math.radians( arcminsperpix[0]/60.0 )                                               # The vertical angular extent of each science pixel

        index = []
        for iv in range(Nv):                                                                                            # for each vertical pixel
            startfovv = bottomangle + deltav * iv                                                                       # define the bottom or start of the vertical pixel
            fov        = FOV_Rectangle(Hrange=(1, 0.0, deltah), Vrange=(1,  startfovv+deltav/2.0, deltav))              # Define a rectangular field of view for the science pixel fields of view for each "wavelength" spectral_window
            idx = self.add_field_of_view(fov)                                                                                 # Add this field of view to our list of field of views
            index.append(idx)

        fovindices = np.array(index)
        return fovindices

    #------------------------------------------------------------------------------
    #           _get_default_list_of_wavelengths
    #------------------------------------------------------------------------------
    def _get_default_list_of_wavelengths(self, wavelen_stepsize):

        waverange = self.specs.wavelength_range_nm
        wavelen = np.arange(waverange[0], waverange[1], wavelen_stepsize)
        return wavelen

    # ------------------------------------------------------------------------------
    #           _make_wavelength_range
    # ------------------------------------------------------------------------------

    def make_wavelength_range(self, wavelengths: np.ndarray = None, wavelen_stepsize=10.0, resolution=1.5) -> np.ndarray:
        """
        Returns an array of wavelengths that span the range of wavelengths used by this ALTIUS channel
        """
        self._wavelen  = np.copy( wavelengths ) if wavelengths is not None else self._get_default_list_of_wavelengths(wavelen_stepsize )
        Nw = self._wavelen.size
        index = []
        for iw in range(Nw):
            idx = self.add_wavelength_range( Wavelength_GaussianPSF(1, self._wavelen[iw], resolution, rtm_processing_scheme='hybrid', photon_distribution='additive'))
            index.append(idx)
        self._waveindices = np.array(index)
        return self._waveindices

    # ------------------------------------------------------------------------------
    #           _make_instrument
    # ------------------------------------------------------------------------------

    def make_instrument(self, max_height_km=90.0, wavelengths = None, wavelen_stepsize=10.0):
        """
        A function to make the spectral windows
        """

        self._fovindices  = self.make_height_profile_of_pixel_fov( max_height_km)                                        # Make the field of view objects each pixel in the height direction covering a height range of max_height_km
        self._waveindices = self.make_wavelength_range( wavelengths=wavelengths, wavelen_stepsize = wavelen_stepsize)   # make the array of wavelength bandpasses
        aperture_area_cm2 = self.specs.aperture_size_cm2
        numfov = self._fovindices.size
        numwav = self._waveindices.size

        for iv in range(numfov):                                                                                        # for each height
            fovidx = self._fovindices[iv]                                                                               # get the index for this field of view
            for iw in range(numwav):                                                                                    # then for all the wavelengths
                wavidx = self._waveindices[iw]                                                                          # get the index of this wavelength bandpass
                self.add_spectral_window(SpectralWindow(self, fovidx, wavidx,aperture_area_cm2))                        # create the spectral window for ALTIUS UV Wavelength 1 at this height
        return self

