from typing import Tuple
import os.path
import numpy as np
import scipy.io
from netCDF4 import Dataset
from .. configuration import OpticsSpecs, AltiusConfigurationKey, SRFUVSpecs, Coords, WindowMappingType


# ------------------------------------------------------------------------------
#           SRFUV
# ------------------------------------------------------------------------------
class SRFUV:
    """
    A class used to manage the ALTIUS UV SRF settings as a fnctin of wavelength.
    This uses settings extracted from the QINETIOQ SPS database and written into a
    Matlab MAT file used by this code snippet.

    I made the file by hand. I ran Lorne's matlab script, `usask_sps/scripts/+usask/+sps/simulateLimbL1Data.m' and set a breakpoint
    at line 282 in `ALTIUS_SPS/QS/+qs/+sps/+m3ism/M3_ISM_loadData.m` just before it executed
    `ALTIUS_instrument.UV_BL_instrument.merge_fields(ALTIUS_instrument.UV_BL_instrument.config)`

    I extracted ALTIUS_instrument.UV_BL_instrument.SRF, SRF_WL, SRF_param, tuneable_WL and tuneable_param and saved
    them to a Matlab save file.

    """

    # ------------------------------------------------------------------------------
    #           __init__
    # ------------------------------------------------------------------------------
    def __init__(self, config: AltiusConfigurationKey):

        self._config = config
        self._optics = OpticsSpecs(config)
        self._specs: SRFUVSpecs = SRFUVSpecs(config)
        self._num_airgaps: int = 0                                  # The number of airgap settings. There is one SRF transmission spectrum for each airgap of each FPI. Typically 1251.
        self._num_srf_wavelen = 0                                   # The number of wavelengths in each SRF transmission spectrum, typically 10001
        self._num_wavelen_settings = 0                              # the number of wavelength settings for the entire FPI stack. Typically 1201
        self._airgap_transmission: np.ndarray = None                # The array of SRF transmission spectra for each airgap of each FPI.
        self._optics_transmission: np.ndarray = None                # The SRF transmission spectra for the UV optics channel. Includes the band-pass filter.
        self._detector_qe: np.ndarray = None                        # The SRF transmission spectra for the UV detector quantum efficiency.
        self._srf_spectra_wavelen: np.ndarray = None                # The Wavelength in nm of the SRF transmisison spectra
        self._tuneable_wavelen: np.ndarray = None                   # The array of wavelengths for which there are FPI airgap settings
        self._current_full_srf: np.ndarray = None
        self._current_fpi: np.ndarray = None
        self._current_central_wavelen: float = 0.0
        self._load_airgap_settings(self._specs.current_entry.fpi_transmission_filename)
        tuneable_name = self._specs.current_entry.tuneable_param_file
        if (tuneable_name is not None):
            self._load_tuneable_params(tuneable_name)

    # ------------------------------------------------------------------------------
    #           _convert_srf_param_to_airgap_settings
    # ------------------------------------------------------------------------------
    def _convert_srf_param_to_airgap_settings(self, tuneable_param: np.ndarray, srf_param: np.ndarray):

        self._num_wavelen_settings = self._tuneable_wavelen.size
        self._wavelen_to_airgap = np.zeros([4, self._num_wavelen_settings], dtype=np.int32)
        for iw, w in enumerate(self._tuneable_wavelen):
            tuneparam = tuneable_param[:, iw].reshape([4, 1])
            srfind = np.argmin(np.abs(srf_param - tuneparam), axis=1)
            self._wavelen_to_airgap[:, iw] = srfind
        self._current_fpi = np.zeros((4, self._srf_spectra_wavelen.size))

    # ------------------------------------------------------------------------------
    #           load_matlab_file
    # ------------------------------------------------------------------------------
    def _load_matlab_file(self, filename):

        srfdata = scipy.io.loadmat(filename)
        print('This function still needs debugging properly')
        SRF = srfdata["SRF"]
        SRF_param = srfdata["SRF_param"]
        SRF_WL = srfdata["SRF_WL"]

        s = SRF[0][0].shape
        self._num_airgaps = s[1]
        self._num_srf_wavelen = s[0]
        self._srf_spectra_wavelen = SRF_WL * 1.0E9                                     # The wavelengths in nm of the select_SRF _srf_spectra_wavelen dimension of the data
        self._tuneable_wavelen = srfdata["tuneable_WL"][:, 0] * 1.0E9
        self._airgap_transmission = np.zeros((4, self._num_airgaps, self._num_srf_wavelen))                              # Array to store the FPI spectral transmission for each airgap. [4, Num airgaps, Num SRF wavelengths]
        self._airgap_transmission[0, :, :] = np.transpose(SRF[0][0][:, :, 0])              # FPI 1, 0, Wavelength setting (1251), SRF wavelength (10001)
        self._airgap_transmission[1, :, :] = np.transpose(SRF[0][1][:, :, 0])              # FPI 2 select_SRF, 2-D array [ select_SRF _srf_spectra_wavelen, tuneable setting]
        self._airgap_transmission[2, :, :] = np.transpose(SRF[0][2][:, :, 0])              # FPI 3 select_SRF, 2-D array [ select_SRF _srf_spectra_wavelen, tuneable setting]
        self._airgap_transmission[3, :, :] = np.transpose(SRF[0][3][:, :, 0])              # FPI 4 select_SRF, 2-D array [ select_SRF _srf_spectra_wavelen, tuneable setting]
        self._optics_transmission = SRF[0][4][:, 0, 0]                                   # UV _optics_transmission select_SRF as function of SRF_wavelength , 1-D array   [select_SRF wavelength]
        self._detector_qe = SRF[0][5][:, 0, 0]                                          # UV Detector select_SRF as a function of SRF_wavelength , 1-D array [select_SRF wavelength]

        s = SRF_param[0][0]
        srf_param = np.zeros([4, s.shape[0]])
        srf_param[0, :] = SRF_param[0][0][:, 0] * 1.0E9
        srf_param[1, :] = SRF_param[0][1][:, 0] * 1.0E9
        srf_param[2, :] = SRF_param[0][2][:, 0] * 1.0E9
        srf_param[3, :] = SRF_param[0][3][:, 0] * 1.0E9
        tuneable_param = np.transpose(srfdata["tuneable_param"] * 1.0E9)
        self._convert_srf_param_to_airgap_settings(tuneable_param, srf_param)

    # ------------------------------------------------------------------------------
    #           _load_sps_oip_file
    # ------------------------------------------------------------------------------
    def _load_sps_oip_file(self, filename):

        with Dataset(filename, 'r') as ds:
            v = ds.variables

            fpi0 = v['SRF01']
            fpi1 = v['SRF02']
            fpi2 = v['SRF03']
            fpi3 = v['SRF04']
            optics = v['SRF05']
            detector = v['SRF06']

            s = fpi0.shape
            self._num_airgaps = s[1]
            self._num_srf_wavelen = s[2]
            self._airgap_transmission = np.zeros((4, self._num_airgaps, self._num_srf_wavelen))                              # Array to store the FPI spectral transmission for each airgap. [4, Num airgaps, Num SRF wavelengths]
            self._airgap_transmission[0, :, :] = fpi0[0, :, :]              # FPI 1, 0, Wavelength setting (1251), SRF wavelength (10001)
            self._airgap_transmission[1, :, :] = fpi1[0, :, :]              # FPI 2 select_SRF, 2-D array [ select_SRF _srf_spectra_wavelen, tuneable setting]
            self._airgap_transmission[2, :, :] = fpi2[0, :, :]              # FPI 3 select_SRF, 2-D array [ select_SRF _srf_spectra_wavelen, tuneable setting]
            self._airgap_transmission[3, :, :] = fpi3[0, :, :]              # FPI 4 select_SRF, 2-D array [ select_SRF _srf_spectra_wavelen, tuneable setting]
            self._optics_transmission = np.squeeze(np.array(optics[0, :, :]))                                     # UV _optics_transmission select_SRF as function of SRF_wavelength , 1-D array   [select_SRF wavelength]
            self._detector_qe = np.squeeze(np.array(detector[0, :, :]))                                           # UV Detector select_SRF as a function of SRF_wavelength , 1-D array [select_SRF wavelength]

            self._srf_spectra_wavelen = np.array(v['SRF_WL'])                                                   # The wavelengths in nm of the select_SRF _srf_spectra_wavelen dimension of the data
            self._tuneable_wavelen = np.array(v["tuneable_WL"]) * 1.0E9
            tuneable_param = np.array(v["tuneable_param"]) * 1.0E9
#            tuneable_param_err = np.array(v["tuneable_param_err"]) * 1.0E09

            s = v['SRF_param01'].shape
            srf_param = np.zeros([4, s[0]])
            srf_param[0, :] = v['SRF_param01'][:] * 1.0E9
            srf_param[1, :] = v['SRF_param02'][:] * 1.0E9
            srf_param[2, :] = v['SRF_param03'][:] * 1.0E9
            srf_param[3, :] = v['SRF_param04'][:] * 1.0E9

            self._convert_srf_param_to_airgap_settings(tuneable_param, srf_param)

    # ------------------------------------------------------------------------------
    #           _load_tuneable_params
    # ------------------------------------------------------------------------------
    def _load_tuneable_params(self, filename):

        raise Exception("This code needs to be updated  and re-written")
        # data = np.loadtxt(filename)
        # self._tuneable_wavelen = data[1:, 0]
        # tuneable_param = np.transpose(data[1:, 1:5])

    # ------------------------------------------------------------------------------
    #           load_airgap_settings
    # ------------------------------------------------------------------------------
    def _load_airgap_settings(self, filename):
        name, ext = os.path.splitext(filename)
        if (ext.lower() == '.m'):
            self._load_matlab_file(filename)
        elif (ext.lower() == '.nc'):
            self._load_sps_oip_file(filename)
        else:
            raise Exception('Unsupported filename')

    # ------------------------------------------------------------------------------
    #           airgap_settings
    # ------------------------------------------------------------------------------
    def airgap_settings(self, target_wavelen_nm) -> Tuple[np.ndarray, float]:
        """
        Gets the array of 4 air-gap settings used to configure the 4 FPI's for a target wavelength.

        Parameters
        ----------
        target_wavelen_nm: float
            The desired target wavelength in nm. It should be in the range 250 to 370.

        Returns
        -------
        Tuple[np.ndarray, float]
            A two element tuple. The first element is an array of 4 airgap settings. The 2nd element is the actual central
            wavelength o fthe airgap settings.
        """

        iw = np.abs(self._tuneable_wavelen - target_wavelen_nm).argmin()                    # Find the tuneable wavelength closest to the desired wavelength
        fpi_airgap_indices = self._wavelen_to_airgap[:, iw]
        return fpi_airgap_indices, self._tuneable_wavelen[iw]

    # ------------------------------------------------------------------------------
    #           select_SRF
    # ------------------------------------------------------------------------------
    def select_SRF(self, central_wavelen_nm: float) -> np.ndarray:
        """
        Select settings and re-calculate the total SRF for the given wavelength. The ALTIUS UV system can only resolve 0.1 nm steps
        between 250 nm and 370.0 nm. This will update the SRF of the 4 constituent FPIs as well as the total SRF
        and the central wavelength. The central wavelength is set to the actual wavelength used from the settings
        database rather than the wavelength requested by the user.

        This code has been compared (2020-07-17) to the QinetiQ code in ALTIUS_SPS\QS\+qs\+sps\+m0generic\+obs\@Instrument\getSRF.m
        in repository git@arggit.usask.ca:qinetiq/sps.git for 310 nm and gave identical answers (for that wavelength).

        Parameters
        ----------
        central_wavelen_nm : float
            The target central wavelength in nm.  It should be a value between 250 and 370 nm at 0.1 nm resolution

        returns
        -------
        np.ndarray
            The complete spectral response function for the wavelength setting closest to the target wavelength.
        """

        fpi_airgap_indices, self._current_central_wavelen = self.airgap_settings(central_wavelen_nm)                         # find the select_SRF parameters closest to this tuneable parameter for each axis
        self._current_full_srf = self._optics_transmission * self._detector_qe

        for i in range(4):
            self._current_fpi[i, :] = self._airgap_transmission[i, fpi_airgap_indices[i], :]
            self._current_full_srf *= self._current_fpi[i, :]
        return self._current_full_srf

    # ------------------------------------------------------------------------------
    #           srf
    # ------------------------------------------------------------------------------
    @property
    def srf(self) -> np.ndarray:
        """
        returns the current total SRF generated from the last call to select_SRF()
        """
        return self._current_full_srf

    # ------------------------------------------------------------------------------
    #           fpi_srf
    # ------------------------------------------------------------------------------
    @property
    def fpi_srf(self) -> np.ndarray:
        """
        Returns the current SRF for the four FPI units as a 2-D array (4,srf_wavelen)
        """
        return self._current_fpi

    # ------------------------------------------------------------------------------
    #           optics_srf
    # ------------------------------------------------------------------------------
    @property
    def optics_srf(self) -> np.ndarray:
        """
        Returns the SRF for the UV optics
        """
        return self._optics_transmission

    # ------------------------------------------------------------------------------
    #           detector_srf
    # ------------------------------------------------------------------------------
    @property
    def detector_srf(self) -> np.ndarray:
        """
        Returns the SRF of the UV detector
        """
        return self._detector_qe

    # ------------------------------------------------------------------------------
    #           srf_wavelen
    # ------------------------------------------------------------------------------
    @property
    def srf_wavelen(self) -> np.ndarray:
        """
        returns the wavelength in nanometers of the SRF spectra in nm
        """

        return self._srf_spectra_wavelen

    # ------------------------------------------------------------------------------
    #           central_wavelength
    # ------------------------------------------------------------------------------
    @property
    def central_wavelength(self) -> float:
        """
        returns the central wavelength set from the last call to select_SRF()
        """
        return self._current_central_wavelen


if __name__ == '__main__':

    import matplotlib.pyplot as plt

    uvsrf = SRFUV()                                                                             # Create the UV SRF object
    uvsrf.select_SRF(310.0)                                                                             # Select an operation wavelength

    plt.figure(1)                                                                                       # plot the total SRF
    plt.plot(uvsrf.srf_wavelen, uvsrf.srf)
    plt.xlabel('Wavelength nm')
    plt.ylabel('SRF')
    plt.title('Total SRF at {:6.2f} nm'.format(uvsrf.central_wavelength))

    plt.figure(2)                                                                                       # plot the constituent SRF components from the 4 FPIs (excludes optics and detector)
    plt.plot(uvsrf.srf_wavelen, uvsrf.fpi_srf.transpose())
    plt.xlim([240, 400])
    plt.xlabel('Wavelength nm')
    plt.ylabel('SRF')
    plt.title('Component SRF at {:6.2f} nm'.format(uvsrf.central_wavelength))

    plt.show()
    print('Done')
