from typing import Tuple, Dict
import os.path
import pickle
import numpy as np
import scipy.signal
import matplotlib.pyplot as plt
from .. configuration import OpticsSpecs, WindowTransforms, AltiusWindowSpecs, WindowMappingHierarchy, AltiusConfigurationKey
from .. psf import PSFTools, NOOFStrayPSF
from .. radiance_simulator.bisa_stray_light_stimulus_data.noof_simulus_reader import SimulationRadianceDataBISA


# ------------------------------------------------------------------------------
#           AltiusStrayLightAnalysis_MeasurementSimulator
# ------------------------------------------------------------------------------
class AltiusStrayLightAnalysis_MeasurementSimulator:
    """
    This is a class that is used to simulate signals measured by ALTIUS. It is a quick replacement for the ALTIUS SPS.
    The class uses "true" simulation data produced by BISA. The class is able to calculate stray light due to infield,
    horizontal NOOF and vertical NOOF as well as the measured signals.

    The class performs FFT convolution to calculate the various stray light signals. This can be computationally onerous
    for VIS and NIR channels, taking several minutes, which can be annoying during software developement and investigation.
    The class implements a simple caching scheme using the wavelength and configuration information.

    Parameters
    ----------
    config: AltiusConfigurationKey
        Configuration info for this simulation. This is used to load ALTIUS Window definitions and PSF images.
    siimulation_data: SimulationRadianceDataBISA
        The simulation data which contain the true radiance signals at the front aperture o fthe instrument.
    """

    def __init__(self, config: AltiusConfigurationKey, simulation_data: SimulationRadianceDataBISA):

        self._configkey: AltiusConfigurationKey = config
        self._simulation: SimulationRadianceDataBISA = simulation_data
        self._cache_dir: str = None
        self._use_cache: bool = True

    # ------------------------------------------------------------------------------
    #           simulation
    # ------------------------------------------------------------------------------
    @property
    def simulation(self) -> SimulationRadianceDataBISA:
        """The Simulation data object"""
        return self._simulation

    def set_cache_dir(self, dirname: str):
        """
        Sets the directory for storing cache data.
        """
        self._cache_dir = dirname
        print('Stray light cached files will be stored and read from folder [{}]'.format(dirname))

    # ------------------------------------------------------------------------------
    #           noof_image_mask_everything_except_vertical_noof
    # ------------------------------------------------------------------------------
    @classmethod
    def _fullconv_image_mask_everything_except_vertical_noof(cls, fullconvimage: np.ndarray, window_specs: AltiusWindowSpecs) -> np.ndarray:

        hierarchy = WindowMappingHierarchy(window_specs)
        sci_window = hierarchy.sci_window.CreateMappingToUpperLevel(hierarchy.noof_window)
        nofsx = hierarchy.noof_window.Ofsx()
        nofsy = hierarchy.noof_window.Ofsy()
        nx = hierarchy.noof_window.Numx()
        ny = hierarchy.noof_window.Numy()
        sofsy = sci_window.Ofsy()
        sy = sci_window.Numy()
        ofsy = nofsy + sofsy

        newimage = np.zeros_like(fullconvimage)
        newimage[nofsy: ofsy, nofsx: nofsx + nx] = fullconvimage[nofsy: ofsy, nofsx: nofsx + nx]
        newimage[ofsy + sy: nofsy + ny, nofsx: nofsx + nx] = fullconvimage[ofsy + sy: nofsy + ny, nofsx: nofsx + nx]
        return newimage

    # ------------------------------------------------------------------------------
    #           noof_image_mask_everything_except_horizontal_noof
    # ------------------------------------------------------------------------------
    @classmethod
    def _fullconv_image_mask_everything_except_horizontal_noof(cls, fullconvimage: np.ndarray, window_specs: AltiusWindowSpecs) -> np.ndarray:

        hierarchy = WindowMappingHierarchy(window_specs)
        sci_window = hierarchy.sci_window.CreateMappingToUpperLevel(hierarchy.noof_window)
        nofsx = hierarchy.noof_window.Ofsx()
        nofsy = hierarchy.noof_window.Ofsy()
        nx = hierarchy.noof_window.Numx()
        sofsx = sci_window.Ofsx()
        sofsy = sci_window.Ofsy()
        sx = sci_window.Numx()
        sy = sci_window.Numy()
        ofsx = nofsx + sofsx
        ofsy = nofsy + sofsy

        newimage = np.zeros_like(fullconvimage)
        newimage[ofsy: ofsy + sy, nofsx: ofsx] = fullconvimage[ofsy: ofsy + sy, nofsx: ofsx]
        newimage[ofsy: ofsy + sy, ofsx + sx: nofsx + nx] = fullconvimage[ofsy: ofsy + sy, ofsx + sx: nofsx + nx]
        return newimage

    # ------------------------------------------------------------------------------
    #           noof_image_mask_science_area
    # ------------------------------------------------------------------------------
    @classmethod
    def _fullconv_image_mask_everything_except_science_area(cls, noofimage: np.ndarray, window_specs: AltiusWindowSpecs) -> np.ndarray:

        hierarchy = WindowMappingHierarchy(window_specs)
        sci_to_fullconv = hierarchy.sci_window.CreateMappingToUpperLevel(hierarchy.fullconv_window)
        ofsx = sci_to_fullconv.Ofsx()
        ofsy = sci_to_fullconv.Ofsy()
        sx = sci_to_fullconv.Numx()
        sy = sci_to_fullconv.Numy()
        newimage = np.zeros_like(noofimage)
        newimage[ofsy:ofsy + sy, ofsx: ofsx + sx] = noofimage[ofsy:ofsy + sy, ofsx: ofsx + sx]
        return newimage

    # ------------------------------------------------------------------------------
    #           make_l1c_stimulus_signals
    # ------------------------------------------------------------------------------

    def make_l1c_stimulus_signals(self, stimulus_example: int, stimulus_wavelen_index: int, do_diagnostics: bool = False) -> Tuple[Dict[str, np.ndarray], np.ndarray, np.ndarray, float]:
        """
        Makes a set of "true" signals using the requested stimulus and wavelength. The various true signals are returned as a dictionary'
        of values. The various true signals are calculated using FFT convolution and can take several minutes for VIS and NIR as the convolution
        must be performed over a fairly large area.

        The trick when making the stray light signals is to always use the measured PSF for infield stray but use a fitted Modified Harvey
        curve when extrapolating over the NOOF window of the VIS and NIR. In this sense the *true* answer is not exactly *true* as we are
        making a tiny approximation. All convolutions for the UV channel use the measured PSF. In this case its not obvious that
        the measured infield PSF in the UV is appropriate for the NOOF convolutions, in-particular the two large ghosts may not be present in the
        true NOOF PSF.

        Parameters
        ----------
        stimulus_example: int
            The index of the stimulus to be processed
        stimulus_wavelen_index: int
            The index of the stimulus wavelength to be processed
        do_diagnostics: bool
            True if diagnostics is required. Default is False.

        Returns
        --------
        Tuple[Dict[str, np.ndarray], np.ndarray, np.ndarray, float]
            Returns a 4 element tuple.
            * [0] Dictionary of true signals. Contains, *true*, *measured*, *infield_stray*, *horiz_stray*, *vertical_stray*, *sci_infield*
            * [1] Array of L1C angles in radians
            * [2] Array of L1C heights in km
            * [3] float value of wavelength
        """

        self._simulation.load_stimulus(stimulus_example)
        basedir = self._cache_dir if self._cache_dir is not None else './'
        picklename = os.path.normpath(os.path.join(basedir, 'cache', 'truth_cache_{:s}_{:1d}{:1d}.pickle'.format(self._configkey.datestr, stimulus_example, stimulus_wavelen_index)))
        if os.path.exists(picklename) and self._use_cache:
            with open(picklename, 'rb') as f:
                obj = pickle.load(f)
        else:
            obj = self._make_l1c_stimulus_signals_implementation(stimulus_example, stimulus_wavelen_index, do_diagnostics=do_diagnostics)          # calculate the BIA signal, convolving over the entire NOOF
            if (self._use_cache):
                basename, name = os.path.split(picklename)
                os.makedirs(basename, exist_ok=True)
                with open(picklename, 'wb') as f:
                    pickle.dump(obj, f)
        return obj

    # ------------------------------------------------------------------------------
    #           make_l1c_stimulus_signals
    # ------------------------------------------------------------------------------
    def _make_l1c_stimulus_signals_implementation(self, stimulus_example: int, stimulus_wavelen_index: int, do_diagnostics: bool = False) -> Tuple[Dict[str, np.ndarray], np.ndarray, np.ndarray, float]:
        """
        This is a function to make true, measured and infield L1C signals from stimulus signals generated by BISA in
        February 2020. The algorithm loads in the stimulus signals over the full near out of field region and convolves
        them with a PSF that is expanded to fit the entire region. The calculation is split into three sections
        (i) convolution of the true infield signal, (ii) convolution of the horizontal near out of field and (iii)
        convoluution of the vertical near out-of-field. The detector sized PSF array as delivered by OIP is used for the
        infield calculation while an extended modified harvey is used for the near out of field regions if they lie
        outside the detecor window bounds (ie vis and nir).

        Returns
        -------
        Tuple[Dict[str, np.ndarray(Ky,Kx)], np.ndarray(Ky), np.ndarray(Ky), float]
            Returns a 4 elements tuple.

            #. Dictionary of true L1C arrays(Ky, Kx) of 6 computed values,  'true', 'measured', 'infield_stray', 'horiz_stray', 'vertical_stray', 'sci_infield'
            #. L1C angles, array(Ky).  Radians from the boresight (+ve upwards, -ve downwards)
            #. L1C altitude, array(Ly), KMs from the sea-level ground
            #. wavelength, float. The wavelength in nanometers of this stimulus wavelength index
        """

        stimulus_example = stimulus_example
        stimuli_widx = stimulus_wavelen_index
        optics = OpticsSpecs(self._configkey)
        transform = WindowTransforms(optics.detector_specs.window_specs)

        psf, center = PSFTools.load_psf(self._configkey)
        psfstray = NOOFStrayPSF(self._configkey)                                                                        # Object that returns the stray light PSF for the entire NOOF region (may be synthetic or measured)
        psftrue_fullconv = transform.detector_to_fullconv(psf, centered_offsets=True)                                   # This is an array with the real PSF but it only covers the detector window
        psfharvey_fullconv = psfstray.fullconv_image()

        true_image, angle = self._simulation.make_fullconv_image(self._configkey, stimulus_index=stimulus_example, wavelength_index=stimuli_widx)
        sci_only = self._fullconv_image_mask_everything_except_science_area(true_image, optics.detector_specs.window_specs)
        horiz_only = self._fullconv_image_mask_everything_except_horizontal_noof(true_image, optics.detector_specs.window_specs)
        vert_only = self._fullconv_image_mask_everything_except_vertical_noof(true_image, optics.detector_specs.window_specs)

        scimapping = transform._hierarchy.sci_window.CreateMappingToUpperLevel(transform._hierarchy.fullconv_window)    # Now extract the angle (and heights) from the stimulus
        l1cmapping = transform._hierarchy.l1c_window.CreateMappingToUpperLevel(transform._hierarchy.sci_window)         # We hacve to go from fullconv coordinates to L1C
        biny = l1cmapping.Biny()                                                                                        # Get the binning o fthe
        ky = l1cmapping.Numy()
        sciangle = angle[scimapping.Ofsy(): scimapping.Ofsy() + scimapping.Numy()]                                      # select the in-field angles
        sciangle = sciangle.reshape([ky, biny])                                                                         # re-bin for L1C
        l1cangle = np.mean(sciangle, axis=1)                                                                            # and get teh mean angle
        l1caltitude = self._simulation.angle_to_altitude(l1cangle)                                                            # then get the altitude of the mean angle
        wavelength = self._simulation.wavelength[stimulus_wavelen_index]

        sci_stray_fullconv = scipy.signal.fftconvolve(sci_only, psftrue_fullconv, mode='same')                              # Use the full PSF for convolving the science window, this has zeros outside the detector window
        infield_stray_fullconv = scipy.signal.fftconvolve(sci_only, psfharvey_fullconv, mode='same')                        # calculate the infield stray light due to harvey scattering
        horiz_stray_fullconv = scipy.signal.fftconvolve(horiz_only, psfharvey_fullconv, mode='same')                        # calculate the horizontal stray light due to harvey scattering
        vert_stray_fullconv = scipy.signal.fftconvolve(vert_only, psfharvey_fullconv, mode='same')                          # calculate the use the extended modifie harvey PSF for vertical contributions
        measured_signal_fullconv = sci_stray_fullconv + horiz_stray_fullconv + vert_stray_fullconv

        if (do_diagnostics):
            horiz_stray2 = scipy.signal.fftconvolve(horiz_only, psftrue_fullconv, mode='same')
            vert_stray2 = scipy.signal.fftconvolve(vert_only, psftrue_fullconv, mode='same')
            plt.figure(2)
            plt.clf()
            xx = np.arange(1500)
            plt.plot(np.log10(measured_signal_fullconv[2720:4220, 3469]), xx, 'g-', label='measured signal')
            plt.plot(np.log10(true_image[2720:4220, 3469]), xx, 'm-', label='true signal')
            plt.plot(np.log10(horiz_stray_fullconv[2720:4220, 3469]), xx, 'c-', label='horiz stray')
            plt.plot(np.log10(vert_stray_fullconv[2720:4220, 3469]), xx, 'r-', label='vertical stray')
            plt.plot(np.log10(sci_stray_fullconv[2720:4220, 3469]), xx, 'g-.', label='infield signal')
            plt.plot(np.log10(horiz_stray2[2720:4220, 3469]), xx, 'c-.')
            plt.plot(np.log10(vert_stray2[2720:4220, 3469]), xx, 'r-.')
            plt.legend()
            plt.xlabel('Vertical Detector Pixel')
            plt.ylabel('Log10 Signal')
            plt.title('Stray light components from extended and regular PSF')

        measured_signal = transform.fullconv_to_l1c(measured_signal_fullconv)
        sci_stray = transform.fullconv_to_l1c(sci_stray_fullconv)
        true_signal = transform.fullconv_to_l1c(true_image)
        horiz_stray = transform.fullconv_to_l1c(horiz_stray_fullconv)
        vert_stray = transform.fullconv_to_l1c(vert_stray_fullconv)
        infield_stray = transform.fullconv_to_l1c(infield_stray_fullconv)

        if (do_diagnostics):
            row = np.arange(501)
            plt.figure()
            plt.clf()
            plt.plot(np.log10(true_signal[:, 5]), row, 'k-', label='True Signal')
            plt.plot(np.log10(sci_stray[:, 5]), row, 'c-', label='Convolved Infield Signal only')
            plt.plot(np.log10(measured_signal[:, 5]), row, 'r-', label='Full Measured Signal')
            plt.xlabel('Log10 Signal')
            plt.ylabel('Science row/height')
            plt.title('Comparison of signals in VIS channel')
            plt.legend()
            print('Done')
        truth = {}
        truth['true'] = true_signal
        truth['measured'] = measured_signal
        truth['infield_stray'] = infield_stray
        truth['horiz_stray'] = horiz_stray
        truth['vertical_stray'] = vert_stray
        truth['sci_infield'] = sci_stray
        return truth, l1cangle, l1caltitude, wavelength
