from typing import Tuple
from abc import ABC, abstractmethod
import math
import pickle
import os
import os.path
import numpy as np
import scipy.optimize
import scipy.special
import scipy.signal
from ..configuration import OpticsSpecs, AltiusConfigurationKey, WindowTransforms
from ..configuration.psf_configuration import HarveyScatteringCoeffs
from ..psf import NOOFStrayPSF


# ------------------------------------------------------------------------------
#           HighAltitudeProxyStrayLightFitter
# ------------------------------------------------------------------------------
class HighAltitudeProxyStrayLightFitter(ABC):
    """
    This is a class for fitting and removing the vertical component of the ALTIUS stray light. The algorithm is based upon
    convolving a near-out-of-field radiance signal with the point spread function and adjusting the radiance until there
    is a good fit with the high altitude proxy. Modelling of the near out-of-field radiance is limited to very simple functions,
    we currently use a constant, mean value.

    This class is only a base class and users must instantiate a drived instance, see :class:`~.HighAltitudeProxyStrayLightFitter_FFTConvolved` for true 1-D PSF calculations
    or :class:`~.HighAltitudeProxyStrayLightFitter_AnalyticPSF` for an analytic modified Harvey PSF.
    """

    def __init__(self,
                 config: AltiusConfigurationKey,
                 harvey_coeffs: HarveyScatteringCoeffs,
                 background_atmosphere_fit_range: Tuple[float, float]=None,
                 high_altitude_proxy_range: Tuple[float, float]=None):

        self._config = config
        self._opticspecs = OpticsSpecs(config)
        self._harvey_coeffs: HarveyScatteringCoeffs = harvey_coeffs
        self._height: np.ndarray = None                                                                                 # Tangent height of each data point in kms.
        self._thetaH: np.ndarray = None                                                                                 # Angle of each data point in radians
        self._K: np.ndarray = None                                                                                      # Integral of PSF across the near out of field region for each angle in the in-field
        self._background_atmosphere_fit_range = background_atmosphere_fit_range                                         # [50.0, 65.0] Fit the initial background atmosphere to values between these heights in kms
        self._high_altitude_proxy_range = high_altitude_proxy_range                                                     # [50.0, 100.0] Fit the high altitude proxy to data between these heights in kms
        self._last_fitted_coeffs: Tuple[float, float, float, float] = None

    @property
    def Vnoof(self):
        """The size of the mean, constant vertical *near-out-of-field stray light*."""
        return self._last_fitted_coeffs[3]

    # ------------------------------------------------------------------------------
    #               def _select_psf_functional_array(self, good):
    # ------------------------------------------------------------------------------
    @abstractmethod
    def _select_psf_functional_array(self, good, l1c_column: int):
        """Selects the PSF array used for convolution. Differes between UV, VIS and NIR"""
        raise Exception('Cannot call base method. It must be overridden')

    # ------------------------------------------------------------------------------
    #           fit_stray_light_curve
    # ------------------------------------------------------------------------------
    def fit_stray_light_curve(self, height: np.ndarray, theta: np.ndarray, signal: np.ndarray, errorterm: np.ndarray, l1c_column: int):
        """
        Fits the stray light curve to the high altitude proxy using general purpose curve fitting algorithm ``scipy.optimize.curve_fit``. This algorithm
        calls a ``callable`` to evaluate the desired function. This is done by this object as it implements special method ``__call__`` to
        make itself a callable.

        The curve fitting fits 4 parameters which are then stored in ``self._last_fitted_coeffs``. The first three parameters are the quadratic background
        fit and the last parameter is the magnitude of the vertical near-out-of-field stray light.  The fitting algorithm has been configured
        to use the ``trf`` method described in curve_fit documentation. This is intentional so we can bound the fitting limits but is an area that can
        be investigated. We have found that unconstrained variation of the parameters has led to the quadratic background atmosphere starting to fit to the
        stray light light.

        Parameters
        ----------
        height: np.ndarray[Ky]
            The 1-D array of height in kms of the center of each L1C pixel.
        theta: np.ndarray[Ky]
            The 1-D array of angle in radians of the center of each L1C pixel from the boresight.
        signal: np.ndarray[Ky]
            The 1-D array of one column of L1C signal. The column is indicated by parameter ``l1c_column``
        errorterm: np.ndarray[Ky]
            The 1-D array of error values on ``signal``
        l1c_column: int
            The column index (0-based) of the L1C profile being processed
        """

        good = (height > self._high_altitude_proxy_range[0]) & (height < self._high_altitude_proxy_range[1]) & np.isfinite(signal) & np.isfinite(errorterm)                       # select valid height range for fitting stray light
        self._height = height
        self._thetaH = theta
        self._K = self._select_psf_functional_array(good, l1c_column)                                     # then set it up

        h = height[good]
        y = signal[good]
        dy = errorterm[good]
        vnoof0 = math.log(np.max(signal))
        a0, b0, c0 = self.initial_estimate_true_signal(height, signal, errorterm)

        bounds = ([-0.02, -0.3, 0.0, 0.0], [0.000001, 0.00, 100.0, 100.0])
        initial_params = a0, b0, c0, vnoof0
        options = {'xtol': 1.49012e-10, 'ftol': 1.49012e-10}
        # pnew, pcov = scipy.optimize.curve_fit( self, h, y, initial_params, sigma=dy, method='lm', **options)
        try:
            pnew, pcov = scipy.optimize.curve_fit(self, h, y, initial_params, sigma=dy, method='trf', bounds=bounds, **options)
        except:
            pnew = [0.0, 0.0, 0.0, 0.0]

        a = pnew[0]
        b = pnew[1]
        c = pnew[2]
        vnoof = pnew[3]
        self._last_fitted_coeffs = a, b, c, vnoof
        gradientat50 = 2.0 * a * 50.0 + b
        # ***** TODO, VERIFY THAT GRADIENT AT 50km is about -0.125

        # print( "{}, {}, {}, {} {}".format( a,b,c, Vnoof, gradientat50))
        # fignum=1
        # if (fignum is not None):
        #     import matplotlib.pyplot as plt
        #     plt.figure(fignum)
        #     plt.clf()
        #     yfit = self(self._height[good], a,b,c, Vnoof)
        #     ytrue = self(self._height[good], a,b,c, 0.0)
        #     #fitted_stray = self.fitted_stray_light(theta)
        #     plt.clf()
        #     plt.plot(np.log10(signal), height, 'k-', label='Total signal')
        #     plt.plot(np.log10(yfit), self._height[good], 'r-', label='Fitted curve')
        #     plt.plot(np.log10(ytrue), self._height[good], 'c-', label='Fitted true')
        #     #plt.plot(np.log10(fitted_stray), height, 'g-', label='Fitted stray')
        #     plt.legend()
        #     plt.ylabel('Height KM')
        #     plt.xlabel('Log10(signal)')
        #     print('DOne')

    # ------------------------------------------------------------------------------
    #           initial_estimate_true_signal
    # ------------------------------------------------------------------------------
    def initial_estimate_true_signal(self,
                                     height: np.ndarray,
                                     signal: np.ndarray, errorterm: np.ndarray) -> Tuple[float, float, float]:
        """
        Makes an initial estimate of the quadratic terms of the background atmosphere. The current implementation only needs to
        be in the ball-park as the final curve-fitting is allowed to adjust the quadratic terms to achieve a better fit.

        Parameters
        ----------
        height: np.ndarray[Ky]
            The 1-D array of heights of the center of the L1C pixels

        signal: np.ndarray[Ky]
            The 1-D array of l1C signals.

        errorterm: np.ndarray[Ky]
            The 1-D array of L1C error values

        Returns
        -------
        Tuple[float, float, float]
            A three element tuple specifying the three coefficients of the quadratic used to represent the log of the background signal.
        """

        good = (height > self._background_atmosphere_fit_range[0]) & (height < self._background_atmosphere_fit_range[1])
        h = height[good]
        y = np.log(signal[good])
        dy = errorterm[good] / signal[good]
        coeffs = np.polyfit(h, y, 1)
        return (0.0, coeffs[0], coeffs[1])

    # ------------------------------------------------------------------------------
    #           stray_light_func
    # ------------------------------------------------------------------------------
    def __call__(self, h: np.ndarray, a: float, b: float, c: float, vnoof: float):

        #  print( "{}, {}, {}, {}".format( a,b,c, Vnoof))
        #  if (Vnoof > 100):
        #   Vnoof = 100.0
        #  if ( a > 0.0):
        #  a = 0.0
        y = np.exp(c + b * h + a * h * h) + math.exp(vnoof) * self._K
        return y

    # ------------------------------------------------------------------------------
    #           calculate_fitted_stray_light
    # ------------------------------------------------------------------------------
    @abstractmethod
    def fitted_stray_light(self, thetah: np.ndarray, icol: int):
        """
        Returns the fitted stray light. This must be implemented in a derived class, see :class:`~.HighAltitudeProxyStrayLightFitter_FFTConvolved`
        and :class:`HighAltitudeProxyStrayLightFitter_AnalyticPSF`.

        Parameters
        ----------
        thetah: np.ndarray
            Array of angles. Each angle is the vertical distance of apoint in radians from the boresight.
        icol: int
            The L1C column being processed.

        Returns
        -------
        The fitted stray light from the the vertical noof at all of the specified angles.
        """
        raise Exception('Cannot call base method. It must be overridden')


# ------------------------------------------------------------------------------
#           HighAltitudeProxyStrayLightFitter
# ------------------------------------------------------------------------------
class HighAltitudeProxyStrayLightFitter_FFTConvolved(HighAltitudeProxyStrayLightFitter):
    """
    This is a class for fitting and removing the vertical component of the ALTIUS stray light. The algorithm uses the
    convolution of the PSF across the entire near out-of-field (rectangular) vertical region. It is based upon
    convolving a near-out-of-field radiance signal with the point spread function and adjusting the radiance until there
    is a good fit with the high altitude proxy. Modelling of the near out-of-field radiance is limited to very simple functions,
    we currently use a constant "mean" value.

    """

    def __init__(self, config: AltiusConfigurationKey, harvey_coeffs: HarveyScatteringCoeffs, **kwargs):

        super().__init__(config, harvey_coeffs, **kwargs)
        self._config = config
        self._straypsf = NOOFStrayPSF(config)
        self._normalized_stray_light_profile: np.ndarray = None                                                         # The 2d L1C array of out-of-field contribution to each L1C pixel
        self._cache_dir = None
    #       self.configure_vertical_stray_light_function(self._opticspecs, self._config.wavelength, self._harvey_coeffs)

    def set_cache_dir(self, cachedir):
        self._cache_dir = cachedir

    # ------------------------------------------------------------------------------
    #           noof_allones_in_vertical_noof
    # ------------------------------------------------------------------------------
    def configure_vertical_stray_light_function(self, optics: OpticsSpecs, wavelength, harvey_coeffs: HarveyScatteringCoeffs, vignetting: np.ndarray = None, use_cache=True):
        """
        Makes a vertical stray light profile by convolving a unit profile with the vertical section of the NOOF window with a Modified Harvey
        PSF across the science window. The resultant convolution across the NOOF window is then reduced to L1C.

        Returns
        -------
        np.ndarray(Ky,Kx)
            A 2-D array containing the L1C vertical stray light profiles. THis profile is automatically set as the
        """

        basedir = self._cache_dir if self._cache_dir is not None else './'
        filename = os.path.join(basedir, 'highaltproxy_stray_light_{:s}_{:7.3f}.pickle'.format(self._config.datestr, wavelength))
        recalculate = not use_cache
        if use_cache:
            ok = os.path.exists(filename)
            if ok:
                with open(filename, 'rb') as f:
                    obj = pickle.load(f)
                    l1c_vert_stray, internal_harvey_coeffs = obj
                ok = (harvey_coeffs.S == internal_harvey_coeffs.S) and (harvey_coeffs.l == internal_harvey_coeffs.l) and (harvey_coeffs.b0 == internal_harvey_coeffs.b0)
            recalculate = not ok

        if recalculate:
            hierarchy = optics.detector_specs.window_hierarchy
            sci_to_fullconv = hierarchy.sci_window.CreateMappingToUpperLevel(hierarchy.fullconv_window)
            Fy = hierarchy.fullconv_window.Numy()
            Fx = hierarchy.fullconv_window.Numx()
            nofsy = optics.detector_specs.window_specs.NOFSy
            nofsx = optics.detector_specs.window_specs.NOFSx
            ofsy = sci_to_fullconv.Ofsy()
            nx = optics.detector_specs.window_specs.NOOFx
            newimage = np.zeros([Fy, Fx])
            newimage[nofsy: ofsy, nofsx: nofsx + nx] = 1.0

            transform = WindowTransforms(optics.detector_specs.window_specs)
            psf_fullconv = self._straypsf.fullconv_image()
            vert_stray_fullconv = scipy.signal.fftconvolve(newimage, psf_fullconv, mode='same')                                  # Use the full PSF for convolving the science window, this has zeros outside the detector window
            l1c_vert_stray = transform.fullconv_to_l1c(vert_stray_fullconv)

            if (use_cache):
                os.makedirs(basedir, exist_ok=True)
                with open(filename, 'wb') as f:
                    obj = l1c_vert_stray, harvey_coeffs
                    pickle.dump(obj, f)

        self._normalized_stray_light_profile = l1c_vert_stray

    # ------------------------------------------------------------------------------
    #               def _select_psf_functional_array(self, good):
    # ------------------------------------------------------------------------------

    def _select_psf_functional_array(self, good, l1c_icol: int):
        if (self._normalized_stray_light_profile is None):
            self.configure_vertical_stray_light_function(self._opticspecs, self._config.wavelength, self._harvey_coeffs)
        return self._normalized_stray_light_profile[good, l1c_icol]

    # ------------------------------------------------------------------------------
    #           calculate_fitted_stray_light
    # ------------------------------------------------------------------------------
    def fitted_stray_light(self, thetah: np.ndarray, icol: int):

        K = np.interp(thetah, self._thetaH, self._normalized_stray_light_profile[:, icol])
        return math.exp(self.Vnoof) * K


# ------------------------------------------------------------------------------
#           HighAltitudeProxyStrayLightFitter_AnalyticPSF
# ------------------------------------------------------------------------------
class HighAltitudeProxyStrayLightFitter_AnalyticPSF(HighAltitudeProxyStrayLightFitter):
    """
    This is a class for analytic fitting and removing the vertical component of the ALTIUS stray light. TThe analytic version
    calculates the PSF integral only in 1-D. It is quick and convenient but more accurate solutioins use the entire 2-D region
    of the near out-of-field vertical region.

    An analytic modified Harvey PSF can be used for true 1-D PSF calculations. However this PSF calculation does not account from
    all angles in the near-out-of-field region. IN the latter case the user can provide a user supplied
    """

    def __init__(self, config: AltiusConfigurationKey, harvey_coeffs: HarveyScatteringCoeffs, **kwargs):

        super().__init__(config, harvey_coeffs, **kwargs)
        self._theta0 = self._opticspecs.fov_angular_size[0].y                                                                   # start angle in radians of lower near out of field
        self._theta1 = self._opticspecs.noof_angular_size[0].y                                                                  # end angle in radians of lower near out of field
        self.integrate_P0 = lambda x, m, l: l * x * scipy.special.hyp2f1(0.5, -m, 1.5, -x * x)                        # Lambda function that Integrates the zero order term of the modified harvey. This is used if user does not provide  a function.
#       self.integrate_P1 = lambda x, m, l : l * l * np.power(1.0 + x * x, m + 1.0)/(2.0 * m + 2.0)                     # Lambda function that Integrates the first order term of the modified harvey, Currently not used
#       self.integrate_P2 = lambda x, m, l : np.power( l * x, 3) *  scipy.special.hyp2f1(1.5, -m, 2.5, -x * x) / 3.0    # Lambda function that integrates the second order term o fthe modified harvey. Currently not used

    # ------------------------------------------------------------------------------
    #               def _select_psf_functional_array(self, good):
    # ------------------------------------------------------------------------------
    def _select_psf_functional_array(self, good, icol: int):
        return self.integrate_psf_across_noof(self._thetaH[good])

    # ------------------------------------------------------------------------------
    #           integrate_psf_across_noof
    # ------------------------------------------------------------------------------
    def integrate_psf_across_noof(self, thetaH: np.ndarray):
        """
        Analytic integration/convolution of the modified harvey psf across the near out of field region for their contribution
        to a set of in-field angles (thetaH) . The integral of
        the modified harvey is given as :math:`\\int B_0\\,\\left[ 1 + \\left(\\frac{\\theta}{l}\\right)^2\\right]^{-\\frac{S}{2}}\\:\\mathrm{d}\\theta = B_0l\\theta\\left[{}_{2}F_1\\left(\\frac{1}{2}; \\frac{S}{2}; \\frac{3}{2}; -\\left(\\frac{\\theta}{l}\\right)^2\\right)\\right]`

        Parameters
        ----------
        thetaH: np.ndarray or real
            The angle in radians of the in-field measurements.

        Returns
        -------
        Returns the integral of the stray light PSF across the near out-of-field, referred to as :math:`K(\\theta_H)` in the documentation.
        :return:
        """

        l = self._harvey_coeffs.l
        M = -self._harvey_coeffs.S / 2.0
        XX0 = (thetaH - self._theta0) / l                                                                               # Get the angle of the start of the near out of field from theta at height h, normalized by shoulder width, l
        XX1 = (thetaH - self._theta1) / l                                                                               # Get the angle of the end of the near out of field from theta at height h, normalized by shoulder width, l
        int0 = self.integrate_P0(XX0, M, l)
        int1 = self.integrate_P0(XX1, M, l)
    #       int0 = XX0 *  scipy.special.hyp2f1( 0.5, -M, 1.5, -XX0 * XX0)                                                   # Evaluate the indefinite integral at the start of the near out-of-field
    #       int1 = XX1 *  scipy.special.hyp2f1( 0.5, -M, 1.5, -XX1 * XX1)                                                   # Evaluate the indefinite integral at the end of the near out-of-field
        K = (int1 - int0) * self._harvey_coeffs.b0                                                                # Evaluate the complte integral across the near out of field. Apply the required  extra terms require
        return K

    # #------------------------------------------------------------------------------
    # #           integrate_psf_across_noof_quadratic_variation
    # #------------------------------------------------------------------------------
    # def integrate_psf_across_noof_quadratic_variation(self, thetaH : np.ndarray, c: float, b: float, a:float):
    #
    #     print("integrate_psf_across_noof_quadratic_variation, TEST THIS CODE")
    #     l = self._harvey_coeffs.l
    #     M = -self._harvey_coeffs.S / 2.0
    #     XX0 = (thetaH - self._theta0) / l                                                                               # Get the angle of the start of the near out of field from theta at height h, normalized by shoulder width, l
    #     XX1 = (thetaH - self._theta1) / l                                                                               # Get the angle of the end of the near out of field from theta at height h, normalized by shoulder width, l
    #     intc = (c + b * thetaH + a * thetaH * thetaH) * (self.integrate_P0(XX1, M, l) - self.integrate_P0(XX0, M, l))
    #     intb = (b + 2.0* a * thetaH) * (self.integrate_P1(XX1, M, l) - self.integrate_P1(XX0, M, l))                                                   # Evaluate the indefinite integral at the end of the near out-of-field
    #     inta = a * (self.integrate_P1(XX1, M, l) - self.integrate_P1(XX0, M, l))
    #     K  =  (intc + intb + inta) * self._harvey_coeffs.b0                                                                # Evaluate the complte integral across the near out of field. Apply the required  extra terms require
    #     return K

    # ------------------------------------------------------------------------------
    #           calculate_fitted_stray_light
    # ------------------------------------------------------------------------------
    def fitted_stray_light(self, thetah: np.ndarray, icol: int):

        assert len(self._last_fitted_coeffs) == 4
        K = self.integrate_psf_across_noof(thetah)
        return math.exp(self.Vnoof) * K
