from typing import Tuple, NamedTuple
import math
import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
import os.path
import scipy.io
import scipy.interpolate
import scipy.signal
from altius_l1pp.altius_deconvolution import AltiusViewportDefn, AltiusViewport, AltiusViewport_UV, AltiusDeconvolutionMatrix
from altius_l1pp.altius_bright_limb_signals import AltiusSimulator_FrontEndRadiance


# ------------------------------------------------------------------------------
#           class AltiusViewport_UV_NOOF
# ------------------------------------------------------------------------------
class AltiusViewport_UV_NOOF(AltiusViewport):

    def __init__(self):
        viewport = AltiusViewportDefn(rawimage_nx=1504,
                                      rawimage_ny=2000,
                                      viewport_x0=163,
                                      viewport_y0=410,
                                      viewport_nx=1176,
                                      viewport_ny=1176,
                                      science_x0=0,
                                      science_y0=0,
                                      science_xbin=3,
                                      science_ybin=3,
                                      science_nx=392,
                                      science_ny=392,
                                      deconv_x0=0,
                                      deconv_y0=0,
                                      deconv_xbin=28,
                                      deconv_ybin=1,
                                      deconv_nx=14,
                                      deconv_ny=392)
        radians_per_raw_pixel = 0.034 / 513.0
        super().__init__(viewport, radians_per_raw_pixel)


# ------------------------------------------------------------------------------
#           load_true_windowed_profile
# ------------------------------------------------------------------------------
def load_true_windowed_profile(basedir: str, filename: str) -> Tuple[np.ndarray, np.ndarray]:
    truedataset = xr.open_dataset( os.path.join(basedir,filename) )
    windowimage = np.array( truedataset.totalRealImageWindowed).transpose()
    y = windowimage.reshape( [171,3,171,3])
    binnedimage = np.sum( y, axis=(1,3))
    return windowimage, binnedimage

# ------------------------------------------------------------------------------
#           load_l1b_image
# ------------------------------------------------------------------------------
def load_l1_image(basedir: str, filename: str) -> Tuple[np.ndarray, np.ndarray]:

    l1bdataset= xr.open_dataset(os.path.join(basedir,filename), group='ModalObservations/Mode 1/RadiometryData')
    l1b = np.squeeze(np.array(l1bdataset.radiometric_signal).transpose())
    l1berror = np.squeeze(np.array(l1bdataset.radiometric_signal_error).transpose())
    return l1b,l1berror

# ------------------------------------------------------------------------------
#           bin_l1b_dataproduct
# ------------------------------------------------------------------------------
def bin_l1b_dataproduct(l1b) -> np.ndarray:
    z = l1b[:,0:-1].reshape([171,5,34])
    l1bbinned = np.mean( z, axis=2)
    return l1bbinned

# ------------------------------------------------------------------------------
#           load_deconvolution_matrix
# ------------------------------------------------------------------------------
def load_deconvolution_matrix(basedir: str, filename: str) -> np.ndarray:
    mdataset= xr.open_dataset(os.path.join(basedir,filename))
    matrix = np.array(mdataset.deconvolution_matrix.transpose())
    return matrix

# ------------------------------------------------------------------------------
#           load_NOOF_UV_deconvolution_matrix
# ------------------------------------------------------------------------------
def load_NOOF_UV_deconvolution_matrix(basedir: str) -> Tuple[np.ndarray, NamedTuple]:
    ast = scipy.io.loadmat( os.path.join(basedir,'l1c_demo_matrix_NOOF_UV.mat'))
    matrix = ast['A']
    return matrix

# ------------------------------------------------------------------------------
#           load_psf_from_filename
# ------------------------------------------------------------------------------
def load_psf( psf_filename):
    psfs = scipy.io.loadmat(psf_filename)
    psf = psfs.get('psf_uv')
    if (psf is None):
        psf = psfs.get('psf_vis')
    factor = np.sum(psf)
    psf /= factor
    return psf

# ------------------------------------------------------------------------------
#           do_convolution
# ------------------------------------------------------------------------------
def do_convolution( image, psf):

    convolvedimage = scipy.signal.fftconvolve( image, psf, mode='same')
    return convolvedimage


noof_viewport = AltiusViewport_UV_NOOF()
std_viewport = AltiusViewport_UV()

deconvolution_matrix = AltiusDeconvolutionMatrix(noof_viewport)
deconvolution_matrix.load_deconvolution_matrix_from_matlab(r'C:\Users\nickl\Documents\Work\software\ARG_Instruments\QinetiQ\usask_sps_cpp\PSF\l1c_demo_matrix_NOOF_UV.mat')

altius_simulator = AltiusSimulator_FrontEndRadiance(noof_viewport,
                                                    vignetting_fov_angles=(-0.039, -0.017, 0.017, 0.039),
                                                    vignetting_fov_values=(0.0000, 1.0, 1.000, 0.000))
wavelen = 350.0
rawimage = altius_simulator.load_front_end_radiance(wavelen, signal_viewport=std_viewport, make_plots=True)             # Generate a scence at the front aperture
#rawimage = altius_simulator.signal_with_random_noise()
truescienceimage = noof_viewport.convert_raw_image_to_science(rawimage)                                                 # Convert this true scence into the true science image
true_l1c_image = noof_viewport.convert_science_image_to_deconv(truescienceimage)                                        # And convert the true science image into the true deconvolved profiles

psf_filename = r'C:\Users\nickl\Documents\Work\software\ARG_Instruments\QinetiQ\ancillary_data\PSF\2020-10-20-PSF\PSF\psf_uv_300nm_EOL.mat'
psf_filename = r'C:\Users\nickl\Documents\Work\software\ARG_Instruments\QinetiQ\ancillary_data\PSF\2020-10-20-PSF\PSF\psf_vis_550nm_EOL.mat'
psf = load_psf(psf_filename)

blurredrawimage = do_convolution( rawimage, psf)                                                                        # Simulate instrument blurring
blurredscienceimage = noof_viewport.convert_raw_image_to_science(blurredrawimage)
blurred_l1c_image = noof_viewport.convert_science_image_to_deconv(blurredscienceimage)

retrieved_l1c_image = deconvolution_matrix.deconvolve_blurred_signal(blurred_l1c_image)
b2 = deconvolution_matrix.convolve_true_signal(true_l1c_image)

plt.figure(10)
plt.clf()
plt.plot(true_l1c_image[:, 7], '-', label='True signal')
plt.plot(retrieved_l1c_image[:, 7], '-.', label='Deconvolved')
plt.plot(blurred_l1c_image[:, 7], '-', label='Blurred PSF')
plt.plot( b2[:,7], '-.', label='Blurred A*X')
plt.ylabel('Signal DN')
plt.title('Signals, NOOF UV {} nm'.format(wavelen))
plt.legend()

diff = np.abs((retrieved_l1c_image[:, 7] - true_l1c_image[:, 7]) / true_l1c_image[:, 7])
plt.figure(11)
plt.clf()
plt.plot(diff, '-')
plt.ylabel('Fractional Difference')
plt.title('Fractional Difference, NOOF UV {} nm'.format(wavelen))

# Examine how much near field stray light is scattering into the science window area
newimage = np.array(rawimage)                                                                                           # Copy the Full, true image
x0 = std_viewport.viewport_definition.viewport_x0                                                                       # Get the science view port bounds in raw detector pixels
x1 = x0 + std_viewport.viewport_definition.viewport_nx
y0 = std_viewport.viewport_definition.viewport_y0
y1 = y0 + std_viewport.viewport_definition.viewport_ny
newimage[y0:y1, x0:x1] = 0.0                                                                                            # Set the science viewport region to zero
blurrednoofimage = do_convolution( newimage, psf)                                                                       # and do the convolutioon
plt.figure(12)
plt.clf()
plt.imshow(blurrednoofimage)
plt.colorbar()
plt.title('UV NOOF scatter into science viewport')

plt.figure(13)
plt.clf()
xvals = list( range( x0+1, x1))
plt.plot( xvals, blurrednoofimage[998,xvals]/rawimage[998,xvals] )
plt.xlabel('Horizontal Science Window Pixel')
plt.ylabel('Fractional signal')
plt.title('UV NOOF Horizontal fractional signal in science window')

plt.figure(14)
plt.clf()
yvals = list( range( y0+1, y1))
plt.plot( yvals, np.log10(rawimage[yvals,750]), label='True signal' )
plt.plot( yvals, np.log10(blurrednoofimage[yvals,750]), label='NOOF signal' )
plt.legend()
plt.xlabel('Vertical Science Window Pixel')
plt.ylabel('Log10(signal)')
plt.title('UV NOOF Vertical signal in science window')

plt.figure(15)
plt.clf()
yvals = list( range( y0+1, y1))
plt.plot( yvals, np.log10(blurrednoofimage[yvals,750]) - np.log10(rawimage[yvals,750]), label='NOOF Fractional signal' )
plt.xlabel('Vertical Science Window Pixel')
plt.ylabel('Log10(Fractional signal)')
plt.title('UV NOOF Vertical fractional signal in science window')

plt.show()
print("Done")