"""
This module calculates the fraction of near-out-of-field stray light removed by a simple high altitude procxy algorithm.
The proxy algorithm calculates stray light by assuming out-of-field stray light intensity is constant at all angles. The
integration of the out-of-field stray light contribution to ion-field altitudes is performed using and analytic integral of
the modified harver BSDF.

Three typical mid-latitude atmospheric profiles are generated for albedo of 0.0, 0.3 and 1.0 at 450, 500, 550, 600 nm. The
profile is extended to evry vertical column of the detector window and convolved with the PSF using a 2D fourier transform.

"""
from typing import Tuple, Dict
import numpy as np
import matplotlib.pyplot as plt
import scipy.signal
import scipy.optimize
import scipy.special
import math
from altius_l1pp_simulator.psf.psf_tools import HarveyScatteringCoeffs
from altius_l1pp_simulator.altius_bright_limb_signals.stray_light_scene_generator import make_vis_extended_scene

#------------------------------------------------------------------------------
#           angle_to_height
#------------------------------------------------------------------------------

def angle_to_height( angles):
    Rs = 6360.0 + 691.0
    Re = 6360.0
    theta0 = math.acos( (Re+50.0)/Rs)
    theta = theta0 - angles
    h = Rs * np.cos(theta) - Re
    bad = h > 100
    h[bad] = np.nan
    return h


# ------------------------------------------------------------------------------
#           stray_light_func
# ------------------------------------------------------------------------------
# g_m : np.ndarray = None
# g_y : np.ndarray = None

def stray_light_func_old( x: np.ndarray ):

    global g_m
    global g_y

    A = x[0]
    ka = -x[1]
    B = x[2]
    kb = -x[3]
    ca = x[4]
    cb = x[5]
    m2 = np.square(g_m)
    v = A * np.exp( g_m * ka + ca * m2) + B * np.exp( g_m * kb + cb * m2 )
    f = np.log(v) - g_y
    return f


#------------------------------------------------------------------------------
#           print_stats
#------------------------------------------------------------------------------

def print_stats(stats, title):
    allwavelen = sorted(stats.keys())
    allheights = sorted(stats[allwavelen[0]].keys())
    print(title)
    s = '            Corrected'
    for w in allwavelen:
        s = '{}        '.format(s)
    s = s + 'Uncorrected'
    print(s)
    s = 'Height  '
    for w in allwavelen:
        s = '{}  {:5.1f}  '.format(s,w)
    for w in allwavelen:
        s = '{}  {:5.1f}  '.format(s,w)
    print(s)
    print('--------------------------------------------------------------')
    for ih in allheights:
        s = '{:5.1f}'.format(ih)
        for w in allwavelen:
            corr_frac,uncorr_frac = stats[w][ih]
            s = '{}, {:7.4f}'.format(s,corr_frac)
        for w in allwavelen:
            corr_frac,uncorr_frac = stats[w][ih]
            s = '{}, {:7.4f}'.format(s,uncorr_frac)
        print(s)
    print()


#------------------------------------------------------------------------------
#           do_analysis
#------------------------------------------------------------------------------

def do_analysis( albedo, wavelen, numbinned, roe_error, fignum, psf_fraction= 1.0) ->Dict[float,float]:

    labelstr = "Wavelength = {}, Albedo = {}, #binned = {}".format(wavelen, albedo, numbinned)

    harvey_coeffs = HarveyScatteringCoeffs(S=2.345813250018345, l=0.001, b0=2.017801183494384e-07*psf_fraction)             # define the modified harvey stray light parameters
    truesignal, stray_psf, anglevert, anglehoriz = make_vis_extended_scene(wavelen,                                         # Make an extended scene
                                                                           albedo,
                                                                           harvey_coeffs=harvey_coeffs,
                                                                           maximum_value=30000.0*numbinned)                  # Get the nominal true signal in electrons

    good = (anglevert > -0.018) & (anglevert < 0.018)                                       # find the near out of field stray light signal,
    oofimage = np.array( truesignal)                                                        # by copying the total scene
    oofimage[good,:] = 0                                                                    # zeroing the in-field components
    oofsignal = scipy.signal.fftconvolve(oofimage, stray_psf, mode='same')                  # and convolving with the stray light stray_psf

    rng = np.random.default_rng()                                                               # Initialize the random number generator
    totalstray =  scipy.signal.fftconvolve(truesignal, stray_psf, mode='same')                  # Find the total stray signal from convolution of true with the stray psf
    totalsignal = truesignal + totalstray                                                       # get the total signal in electrons
    errorterm = np.sqrt(totalsignal + roe_error * numbinned)                                    # get the error in the total signal
    noisysignal = rng.normal(totalsignal, errorterm)                                            # Get the noisy total signal.

    colidx = 752
    thetah = anglevert[good]
    h = angle_to_height(thetah)
    y = noisysignal[good,colidx]
    dy = errorterm[good, colidx]
    stray_light_fitter = StrayLightFunc(harvey_coeffs)
    stray_light_fitter.fit_stray_light_curve(h, thetah, y, dy, fignum=fignum)
    plt.grid(True)
    plt.title('Fit Results, {}'.format(labelstr))

    fitted_stray = stray_light_fitter.calculate_fitted_stray_light(thetah)
    totalsig = noisysignal[good, colidx]
    truesig  = truesignal[good, colidx]
    corrsig  = totalsig - fitted_stray
    deltasig = (corrsig - truesig) / truesig
    totalfrac = (totalsig - truesig) / truesig

    hrange = np.arange( 5, 70, 5)
    srd_stats = {}
    for ih in hrange:
        idx = (h >= (ih - 2.5)) & (h < (ih + 2.5))
        srd_stats[ih] = np.nanmean( deltasig[idx]), np.nanmean( totalfrac[idx] )

    plt.figure(fignum+1)
    plt.clf()
    plt.plot( np.log10(totalsig), h, 'k-', label='True + Stray Signal')
    plt.plot( np.log10(corrsig), h, 'r-', label='Fitted True Signal')
    plt.plot( np.log10(truesig), h, 'c-', label='True Signal')
    plt.grid(True)
    plt.legend()
    plt.title('Corrected Signals. {}'.format(labelstr))

    plt.figure(fignum+2)
    plt.clf()
    plt.plot()
    plt.plot( deltasig*100.0, h, 's', label='Corrected Signal', markerfacecolor='#00000060', markeredgecolor='#000000a0' )
    plt.plot( totalfrac*100.0, h, 'o', label='Measured Signal', markerfacecolor='#ff000060', markeredgecolor='#ff0000a0' )
    plt.grid(True)
    plt.xlabel('Percentage difference')
    plt.ylabel('Height Kms')
    plt.title('Percentage Difference %. {}'.format(labelstr))
    plt.xlim([-10., 10.0])
    plt.legend()
    return srd_stats



fignum = 20
#albedo = 0.3
#srf_wavelen= 550.0
roe_error = 2.0
allwavelen = 450.0, 500.0, 550.0, 600.0

l1c_albedo1p0_stats = {}
l1c_albedo0p3_stats = {}
l1c_albedo0p0_stats = {}
l1b_albedo1p0_stats = {}
l1b_albedo0p3_stats = {}
l1b_albedo0p0_stats = {}

psf_fractions = [0.5, 0.25, 0.10, 1.0]
for psf_fraction in psf_fractions:
    for i, wavelen in enumerate(allwavelen):
        print("Processing wavelength {} PSF Fraction {:4.2f}".format(wavelen, psf_fraction))
        l1c_albedo1p0_stats[wavelen] = do_analysis(1.0, wavelen, 450, roe_error, 6*i + 1, psf_fraction=psf_fraction)
        l1c_albedo0p3_stats[wavelen] = do_analysis(0.3, wavelen, 450, roe_error, 10)
        l1c_albedo0p0_stats[wavelen] = do_analysis(0.0, wavelen, 450, roe_error, 15)
        l1b_albedo1p0_stats[wavelen] = do_analysis(1.0, wavelen, 9, roe_error, 25)
        l1b_albedo0p3_stats[wavelen] = do_analysis(0.3, wavelen, 9, roe_error, 30)
        l1b_albedo0p0_stats[wavelen] = do_analysis(0.0, wavelen, 9, roe_error, 35)


    print_stats(l1c_albedo1p0_stats, 'L1C Albedo = 1.0, psf fraction = {:4.2f}'.format(psf_fraction))
    print_stats(l1c_albedo0p3_stats, 'L1C Albedo = 0.3, psf fraction = {:4.2f}'.format(psf_fraction))
    print_stats(l1c_albedo0p0_stats, 'L1C Albedo = 0.0, psf fraction = {:4.2f}'.format(psf_fraction))
    print_stats(l1b_albedo1p0_stats, 'L1B Albedo = 1.0, psf fraction = {:4.2f}'.format(psf_fraction))
    print_stats(l1b_albedo0p3_stats, 'L1B Albedo = 0.3, psf fraction = {:4.2f}'.format(psf_fraction))
    print_stats(l1b_albedo0p0_stats, 'L1B Albedo = 0.0, psf fraction = {:4.2f}'.format(psf_fraction))
print('Done')