from typing import Tuple
import numpy as np


# ------------------------------------------------------------------------------
#          Class AltiusStrayLightSystemRequirements
# ------------------------------------------------------------------------------
class AltiusStrayLightSystemRequirements:
    """
    This is a small class that stores the ALTIUS stray light system requirements. It is intended to be for convenience
    more than anyything else
    """

    def __init__(self):

        self._wavelen = np.array([300.0, 350.0, 450.0, 500.0, 550.0, 600.0, 650.0, 700.0])
        self._heights = np.array([10.0, 15.0, 20.0, 25.0, 30.0, 35.0, 40.0, 45.0, 50.0, 55.0, 60.0, 65.0])
        self._srd = np.zeros([self._heights.size, self._wavelen.size])
        self._srd[:, 0] = np.array([0.054, 0.053, 0.052, 0.051, 0.049, 0.047, 0.045, 0.041, 0.039, 0.047, 0.068, 0.115])         # 300 nm
        self._srd[:, 1] = np.array([0.003, 0.003, 0.003, 0.003, 0.004, 0.005, 0.008, 0.010, 0.014, 0.020, 0.030, 0.048])         # 350 nm
        self._srd[:, 2] = np.array([0.001, 0.001, 0.001, 0.001, 0.001, 0.002, 0.002, 0.002, 0.002, 0.003, 0.003, 0.004])         # 450 nm
        self._srd[:, 3] = np.array([0.001, 0.001, 0.001, 0.002, 0.002, 0.002, 0.003, 0.003, 0.004, 0.005, 0.005, 0.006])         # 500 nm
        self._srd[:, 4] = np.array([0.002, 0.002, 0.002, 0.003, 0.004, 0.005, 0.006, 0.007, 0.008, 0.010, 0.012, 0.014])         # 550 nm
        self._srd[:, 5] = np.array([0.003, 0.004, 0.005, 0.006, 0.008, 0.010, 0.012, 0.016, 0.020, 0.026, 0.034, 0.043])         # 600 nm
        self._srd[:, 6] = np.array([0.003, 0.004, 0.005, 0.007, 0.009, 0.011, 0.015, 0.019, 0.024, 0.031, 0.040, 0.051])         # 650 nm
        self._srd[:, 7] = np.array([0.004, 0.005, 0.007, 0.009, 0.011, 0.015, 0.019, 0.024, 0.031, 0.040, 0.052, 0.066])         # 700 nm

    # ------------------------------------------------------------------------------
    #           srd
    # ------------------------------------------------------------------------------
    def srd(self, userwavelen: float) -> Tuple[np.ndarray, np.ndarray]:
        """
        Returns the height profile of ALTIUS stray light system requirements at this wavelength.

        Parameters
        ----------
        userwavelen: float
            The wavelength in nm at which  stray light requirements are required.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            Returns a 2 element tuple. Each element contains a 1-D array. The first element is the height in km of the
            of the system requirement value. The second element of the tuple is the array of system requirement values
            with height.
        """

        s = np.zeros_like(self._heights)
        for ih in range(self._heights.size):
            s[ih] = np.interp(userwavelen, self._wavelen, self._srd[ih, :], left=np.nan, right=np.nan)
        return self._heights, s
